# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""WOQ save and load."""
# pylint:disable=import-error

import copy
import json
import os
import re
import tempfile

import torch
from packaging.version import parse

from neural_compressor.common.utils import AWQ, TEQ, save_config_mapping
from neural_compressor.torch.utils import (
    HPU_SAFE_WEIGHTS_NAME,
    HPU_WEIGHT_NAME,
    LM_HEAD_NAMES,
    QCONFIG_NAME,
    SHARDED_HPU_WEIGHT_NAME,
    SHARDED_WEIGHT_NAME,
    WEIGHT_NAME,
    SaveLoadFormat,
    get_accelerator,
    get_enum_from_format,
    logger,
    set_module,
)

from .modules import HPUWeightOnlyLinear, INCWeightOnlyLinear, MulLinear
from .utility import convert_dtype_str2torch

format_woqlinear_mapping = {
    SaveLoadFormat.HUGGINGFACE: INCWeightOnlyLinear,
    SaveLoadFormat.DEFAULT: INCWeightOnlyLinear,
}
device_woqlinear_mapping = {"cpu": INCWeightOnlyLinear, "hpu": HPUWeightOnlyLinear}

MAX_FILE_SIZE = 5  # GB
cur_accelerator = get_accelerator()


def save(model, output_dir="./saved_results", format=SaveLoadFormat.DEFAULT, **kwargs):
    """Save the quantized model and config to the output path.

    Args:
        model (torch.nn.module): raw fp32 model or prepared model.
        output_dir (str, optional): output path to save.
        format (str, optional): The format in which to save the model. Options include "default" and "huggingface". Defaults to "default".
        kwargs: Additional arguments for specific formats. For example:
            - safe_serialization (bool): Whether to use safe serialization when saving (only applicable for 'huggingface' format). Defaults to True.
            - tokenizer (Tokenizer, optional): The tokenizer to be saved along with the model (only applicable for 'huggingface' format).
            - max_shard_size (str, optional): The maximum size for each shard (only applicable for 'huggingface' format). Defaults to "5GB".
    """
    format = get_enum_from_format(format)
    os.makedirs(output_dir, exist_ok=True)
    cur_accelerator.synchronize()
    if format == SaveLoadFormat.HUGGINGFACE:  # pragma: no cover
        quantization_config_file = "quantize_config.json"
        safe_serialization = kwargs.get("safe_serialization", True)
        max_shard_size = kwargs.get("max_shard_size", f"{MAX_FILE_SIZE}GB")
        if not hasattr(model.config, "quantization_config"):
            quantization_config = change_config_to_hf_format(model.qconfig)
            model.config.quantization_config = quantization_config
        # save model state_dict and config.json
        model.save_pretrained(output_dir, max_shard_size=max_shard_size, safe_serialization=safe_serialization)
        # save quantize_config.json
        with open(os.path.join(output_dir, quantization_config_file), "w", encoding="utf-8") as f:
            json.dump(quantization_config, f, indent=2)
        # save generation_config.json
        if hasattr(model, "generation_config") and model.generation_config is not None:
            model.generation_config.save_pretrained(output_dir)
        # save tokenizer
        tokenizer = kwargs.get("tokenizer", None)
        if tokenizer is not None:
            tokenizer.save_pretrained(output_dir)
        return
    elif format == SaveLoadFormat.DEFAULT:
        output_folder = os.path.abspath(os.path.expanduser(output_dir))
        qmodel_weight_file_path = os.path.join(output_folder, WEIGHT_NAME)
        qconfig_file_path = os.path.join(output_folder, QCONFIG_NAME)
        # saving process
        save_config_mapping(model.qconfig, qconfig_file_path)
        if "blockwise" in kwargs:
            from neural_compressor.torch.algorithms.layer_wise import LWQ_WORKSPACE, save_layers_in_shards_iteratively

            checkpoints_folder = kwargs.get("blockwise_load_folder", None)
            if not checkpoints_folder:
                checkpoints_folder = LWQ_WORKSPACE
            save_layers_in_shards_iteratively(checkpoints_folder, output_folder, layers_per_shard=8)
        else:
            model_state_dict = model.state_dict()  # if 'model_state_dict' not in kwargs else kwargs['model_state_dict']
            torch.save(model_state_dict, qmodel_weight_file_path)
            logger.info("Save quantized model weight to {}.".format(qmodel_weight_file_path))
        logger.info("Save configuration of quantized model to {}.".format(qconfig_file_path))


def load(model_name_or_path, original_model=None, format=SaveLoadFormat.DEFAULT, device="cpu", **kwargs):
    """Load quantized weight-only quantization model.

    1. Load INC weight-only quantized model in local.
        from neural_compressor.torch.quantization import load
        load(model_name_or_path="saved_results", original_model=fp32_model, format="default", device="cpu")

    2. Load HuggingFace weight-only quantized model, including GPTQ models and
       upstreamed INC quantized models in HF model hub.
        from neural_compressor.torch.quantization import load
        load(model_name_or_path=model_name_or_path, format="huggingface", device="cpu")

    Args:
        model_name_or_path (str):  torch checkpoint directory or hugginface model_name_or_path.
            If 'format' is set to 'huggingface', it means the huggingface model_name_or_path.
            If 'format' is set to 'default', it means the 'checkpoint_dir'.
            Parameter should not be None. it coworks with 'original_model' parameter to load INC
            weight-only quantized model in local.
        original_model (torch.nn.module, optional): original model before quantization.
            Needed if 'format' is set to 'default'. Defaults to None.
        format (str, optional): 'defult' for loading INC weight-only quantized model.
            'huggingface' for loading huggingface WOQ causal language model. Defaults to "default".
        kwargs (remaining dictionary of keyword arguments, optional):
            remaining dictionary of keyword arguments for loading huggingface models.
            will be passed to the huggingface model's `__init__` method, such as 'trust_remote_code', 'revision'.

    Returns:
        torch.nn.Module: quantized model
    """
    format = get_enum_from_format(format)
    model_loader = WOQModelLoader(model_name_or_path, original_model, format, device, **kwargs)
    model = model_loader.load_woq_model()
    return model


class WOQModelLoader:
    """WOQ Model Loader."""

    def __init__(self, model_name_or_path, original_model=None, format=SaveLoadFormat.DEFAULT, device="cpu", **kwargs):
        """Init the WOQModelLoader object."""
        self.model_name_or_path = model_name_or_path
        self.original_model = original_model
        self.format = format
        self.device = device
        self.kwargs = kwargs
        self.quantization_config = {}
        self.loaded_state_dict = {}
        self.loaded_state_dict_keys = []
        self._should_save_hpu_format_tensor = False
        self._model_local_dir = None  # local directory where model files are saved
        self.sharded_checkpoints = kwargs.get("sharded_checkpoints", False)

    def _post_process_for_w4a8(self, woq_model):
        # Replace the WoQ Linear with the mixed precision Linear
        from neural_compressor.torch.algorithms.mixed_low_precision.modules import (
            replace_hpu_woq_with_hpu_mixed_precision_linear,
        )

        w4a8_model = replace_hpu_woq_with_hpu_mixed_precision_linear(woq_model)
        from neural_compressor.torch.utils import get_accelerator

        w4a8_model = w4a8_model.to(get_accelerator().name())
        return w4a8_model

    def load_woq_model(self):
        """Load quantized weight-only quantization model.

        Raises:
            ValueError: `format` in load function can only be 'huggingface' or 'default'.

        Returns:
            torch.nn.Module: quantized model
        """
        if self.format == SaveLoadFormat.HUGGINGFACE:
            assert self.model_name_or_path is not None, "'model_name_or_path' can't be None."

            model = self.load_hf_format_woq_model()
            logger.info("Loading HuggingFace weight-only quantization model successfully.")
        elif self.format == SaveLoadFormat.DEFAULT:
            assert os.path.exists(self.model_name_or_path), f"'{self.model_name_or_path}' path doesn't exist."
            assert (
                self.original_model is not None
            ), "Can't get original model. Please pass `original_model` to load function."

            model = self.load_inc_format_woq_model()
            logger.info("Loading weight-only quantization model successfully.")
        else:
            raise ValueError(f"`format` in load function can only be 'huggingface' or 'default', but get {self.format}")

        if self._is_w4a8_model_from_auto_round():
            model = self._post_process_for_w4a8(model)
        if self._is_w4a8_model_from_dpq(model):
            model.dpq_quantized = True

        return model

    def load_inc_format_woq_model(self):
        """Load WOQ model saved in INC file format."""
        self._model_local_dir = self.model_name_or_path

        qmodel_weight_file_path = os.path.join(
            os.path.abspath(os.path.expanduser(self.model_name_or_path)), WEIGHT_NAME
        )
        # if hpu format tensor can be used directly, then update qmodel_weight_file_path to the hpu format tensor file
        if self._use_hpu_module():
            print(f"sharded_checkpoints = {self.sharded_checkpoints}")
            if self.sharded_checkpoints:
                qmodel_weight_file_path = os.path.join(
                    os.path.abspath(os.path.expanduser(self.model_name_or_path)), SHARDED_HPU_WEIGHT_NAME
                )
            else:
                qmodel_weight_file_path = os.path.join(
                    os.path.abspath(os.path.expanduser(self.model_name_or_path)), HPU_WEIGHT_NAME
                )
        else:
            if self.sharded_checkpoints:
                qmodel_weight_file_path = os.path.join(
                    os.path.abspath(os.path.expanduser(self.model_name_or_path)), SHARDED_WEIGHT_NAME
                )
        assert os.path.exists(qmodel_weight_file_path), (
            "Cannot load model weight from path {}. "
            "Please make sure '{}' file is saved in your '{}' directory ".format(
                qmodel_weight_file_path, WEIGHT_NAME, self.model_name_or_path
            )
        )
        logger.info(f"Find weight file {qmodel_weight_file_path}")

        qconfig_file_path = os.path.join(os.path.abspath(os.path.expanduser(self.model_name_or_path)), QCONFIG_NAME)
        assert os.path.exists(qconfig_file_path), (
            "Cannot load model quantization config from path {}. "
            "Please make sure '{}' file is saved in your '{}' directory".format(
                qconfig_file_path, QCONFIG_NAME, self.model_name_or_path
            )
        )

        # get loaded state_dict
        if self.sharded_checkpoints:
            print("loading sharded checkpoints")
            if self._use_hpu_module():
                from safetensors.torch import load_file

                self.loaded_state_dict = load_file(qmodel_weight_file_path, device=self.device)
            else:
                from neural_compressor.torch.algorithms.layer_wise import load_model_from_shards_with_safetensors

                shard_dir = os.path.abspath(os.path.expanduser(self.model_name_or_path))
                bin_index_file = os.path.join(shard_dir, "model_bin_index.json")
                self.loaded_state_dict = load_model_from_shards_with_safetensors(shard_dir, bin_index_file)
        else:
            self.loaded_state_dict = torch.load(qmodel_weight_file_path)

        self.loaded_state_dict_keys = list(set(self.loaded_state_dict.keys()))

        # get qconfig
        with open(qconfig_file_path, "r") as file:
            self.quantization_config = json.load(file)

        # build weight-only quantization model with WeightOnlyLinear module
        print("Building WOQ model")
        model = self._build_woq_model()

        # load remaining pretrained weight to weight-only quantization model
        is_meta_device = hasattr(self.original_model, "device") and self.original_model.device.type == "meta"
        algo_name = next(iter(self.quantization_config[next(iter(self.quantization_config))].keys()))
        if is_meta_device or algo_name in [AWQ, TEQ]:
            # AWQ and TEQ will update some weight except WOQLinear to handle additional input_scale
            model.load_state_dict(self.loaded_state_dict, assign=True, strict=False)

        # save hpu format tensor to local directory
        if self._should_save_hpu_format_tensor:
            print("saving optimized hpu model")
            self._save_hpu_format_tensor(model)

        model.eval()
        return model

    def _is_w4a8_model_from_auto_round(self):
        if self.quantization_config.get("data_type", None) == "fp8_to_int_sym":
            return True
        for layer_config in self.quantization_config.get("extra_config", {}).values():
            if layer_config.get("data_type", None) == "fp8_to_int_sym":
                return True
        return False

    # check if the model was quantized using Dual Precision Quantization (DPQ)
    def _is_w4a8_model_from_dpq(self, model):
        for name, _ in model.named_buffers():
            if "scale_bf16_to_fp8" in name:
                return True
        return False

    def _update_quant_config_for_w4a8(self):
        self.quantization_config["quant_method"] = "gptq"
        self.quantization_config.pop("backend", None)

    def _is_autoround_format_quantized_model(self):
        return "backend" in self.quantization_config and "auto_round" in self.quantization_config["backend"]

    def load_hf_format_woq_model(self):
        """Load HuggingFace weight-only quantized model.

        Returns:
            torch.nn.Module: quantized model
        """
        # check required package
        from neural_compressor.torch.utils import is_package_available

        if not is_package_available("transformers"):
            raise ImportError("Loading huggingface model requires transformers: `pip install transformers`")
        if not is_package_available("accelerate"):
            raise ImportError("Loading huggingface model requires accelerate: `pip install accelerate`")

        # get model class and config
        model_class, config = self._get_model_class_and_config()
        self.quantization_config = config.quantization_config if hasattr(config, "quantization_config") else None
        if self._is_w4a8_model_from_auto_round():
            self._update_quant_config_for_w4a8()
        if self._is_autoround_format_quantized_model():
            # pragma: no cover
            # load autoround format quantized model
            from auto_round import AutoRoundConfig

            hf_kargs = {}
            pretrain_args = ["trust_remote_code", "_attn_implementation", "device_map", "torch_dtype"]
            for item in pretrain_args:
                arg_value = self.kwargs.get(item, None)
                if arg_value is not None:
                    hf_kargs[item] = arg_value
            model = model_class.from_pretrained(self.model_name_or_path, **hf_kargs)
            return model
        # get loaded state_dict
        self.loaded_state_dict = self._get_loaded_state_dict(config)
        self.loaded_state_dict_keys = list(set(self.loaded_state_dict.keys()))

        # initiate the huggingface model (FP32 empty model)
        self.original_model = self._init_hf_model(model_class, config)

        # build weight-only quantization model with WeightOnlyLinear module
        # and load quantized weight to WeightOnlyLinear modules
        model = self._build_woq_model()

        # clear loaded_state_dict
        self.loaded_state_dict = {}

        # load remaining pretrained weight to weight-only quantization model
        model = self._load_remaining_pretrained_weight(model)

        # save hpu format tensor to local directory
        if self._should_save_hpu_format_tensor:
            self._save_hpu_format_tensor(model)

        model.eval()
        return model

    def _is_hqq_model(self):
        for name, module in self.original_model.named_modules():
            pattern = rf"(\(.*{re.escape(name)}.*{re.escape(type(module).__name__)}.*\))"
            for q_config_key, q_config_value in self.quantization_config.items():
                if re.search(pattern, q_config_key):
                    if isinstance(q_config_value, dict) and [algo for algo in q_config_value.keys()][0] == "hqq":
                        return True

    def _build_hqq_model(self):
        """Replace quantized Linear with HQQLinear."""
        from neural_compressor.torch.algorithms.weight_only.hqq.core import HQQLinear
        from neural_compressor.torch.utils import set_module

        for name, module in self.original_model.named_modules():
            if isinstance(module, torch.nn.Linear):
                loaded_state_dict_keys_set = set(self.loaded_state_dict_keys)
                if name + ".val" not in loaded_state_dict_keys_set:
                    continue
                new_module = HQQLinear(
                    in_features=module.in_features, out_features=module.out_features, bias=module.bias is not None
                )
                self._load_data_to_new_module_hqq(new_module, name)
                set_module(self.original_model, name, new_module)
        woq_model = self.original_model
        return woq_model

    def _load_data_to_new_module_hqq(self, new_module, module_name):
        new_module_state_dict = {}
        for key in self.loaded_state_dict:
            if key.startswith(module_name):
                new_key = key[len(module_name) + 1 :]  # Remove module_name and the following dot
                new_module_state_dict[new_key] = self.loaded_state_dict[key]
                self.loaded_state_dict_keys.remove(key)
        new_module.load_state_dict(new_module_state_dict, strict=False)

    def _build_woq_model(self):
        """Build weight-only quantization model."""
        if self._is_hqq_model():
            return self._build_hqq_model()

        self._update_format_woqlinear_mapping()

        for name, module in self.original_model.named_modules():
            # replace `torch.nn.Linear` to `WeightOnlyLinear` in self.original_model and load its quantized data
            if isinstance(module, torch.nn.Linear):
                # module without qweight means it is not quantized, then skip it
                if (
                    name + ".qweight" not in self.loaded_state_dict_keys
                    and name + ".linear.qweight" not in self.loaded_state_dict_keys
                ):
                    continue

                module_quantization_config, _is_autoround = self._get_module_quantization_config(name, module)
                self._replace_woqlinear_modules(name, module, module_quantization_config, _is_autoround)

        woq_model = self.original_model
        return woq_model

    def _update_format_woqlinear_mapping(self):
        """Update format mapping module to HPUWeightOnlyLinear if tensor is hpu format."""
        if self._use_hpu_module():
            format_woqlinear_mapping.update({self.format: HPUWeightOnlyLinear})
        else:  # reset format mapping module to INCWeightOnlyLinear
            format_woqlinear_mapping.update({self.format: INCWeightOnlyLinear})

        logger.debug(
            f"Build weight-only quantization model according to format and device mapping. \n"
            f"Format mapping is {format_woqlinear_mapping}. \n"
            f"Device mapping is {device_woqlinear_mapping}."
        )

    def _get_module_quantization_config(self, module_name, module):
        """Gt quantization config of current module.

        1. INC weight-only quantization model, quantization_config will be structured in module level like:
            {(module1_name, module1_type): {"rtn": {"bits": 4, ...}}, ...}
        2. HF weight-only quantization model, quantization_config will be structured in model level like:
            {'bits': 4, ...}
        """
        module_quantization_config = self.quantization_config
        pattern = rf"(\(.*{re.escape(module_name)}.*{re.escape(type(module).__name__)}.*\))"
        _is_autoround = False
        # for loop is used to find quantization config of the target module in INC weight-only quantization model
        for q_config_key, q_config_value in self.quantization_config.items():
            if re.search(pattern, q_config_key):
                # pattern will map (module_name, moduele_type)
                if isinstance(q_config_value, dict) and [algo for algo in q_config_value.keys()][0] == "autoround":
                    _is_autoround = True
                module_quantization_config = [config for config in q_config_value.values()][0]
        return module_quantization_config, _is_autoround

    def _replace_woqlinear_modules(self, name, linear_module, module_quantization_config, _is_autoround):
        """Replace torch.nn.Linear modules with WeightOnlyLinear and load its quantized data."""
        # insert MulLinear module for AWQ/TEQ algorithm
        if name + ".linear.qweight" in self.loaded_state_dict_keys:
            new_module = MulLinear(linear_module)
            set_module(self.original_model, name, new_module)
            name += ".linear"

        # get format mapping module class
        WeightOnlyLinearClass = format_woqlinear_mapping[self.format]

        # update initialization kwargs for woq linear module
        module_kwargs = {}

        # base initialization kwargs
        module_kwargs["in_features"] = linear_module.in_features
        module_kwargs["out_features"] = linear_module.out_features
        module_kwargs["dtype"] = module_quantization_config.get("dtype", "int")
        module_kwargs["bits"] = module_quantization_config.get("bits", 4)
        module_kwargs["group_size"] = module_quantization_config.get("group_size", 32)

        # specific initialization kwargs
        module_kwargs["g_idx"] = module_quantization_config.get("desc_act", False)
        module_kwargs["zp"] = True if name + ".qzeros" in self.loaded_state_dict_keys else False
        module_kwargs["use_optimum_format"] = True
        module_kwargs["bias"] = linear_module.bias is not None
        if _is_autoround:
            module_kwargs["scale_dtype"] = convert_dtype_str2torch(
                module_quantization_config.get("scale_dtype", "fp16")
            )

        if self._is_w4a8_model_from_auto_round():
            module_kwargs["enable_w4a8"] = True

        # initialize the new WeightOnlyLinearClass
        new_module = WeightOnlyLinearClass(**module_kwargs)

        # load quantized data of current module
        self._load_data_to_new_module(new_module, name)

        # update mapped woqlinear module if needed
        new_module = self._update_mapped_woqlinear_modules(name, new_module, module_kwargs)

        # [SW-234528]: if g_idx is not None, then check whether the g_idx is ordered
        if isinstance(new_module, HPUWeightOnlyLinear) and new_module.g_idx is not None:
            # if g_idx is ordered, then set g_idx to None
            if new_module.is_g_idx_ordered(new_module.g_idx, new_module.group_size):
                setattr(new_module, "g_idx", None)

        set_module(self.original_model, name, new_module)

    def _load_data_to_new_module(self, new_module, module_name):
        new_module_state_dict = {}
        for key in [".qweight", ".scales", ".scale_bf16_to_fp8", ".qzeros", ".bias", ".g_idx"]:
            full_name = module_name + key
            if full_name in self.loaded_state_dict:
                new_module_state_dict[key[1:]] = self.loaded_state_dict.pop(full_name)
                self.loaded_state_dict_keys.remove(full_name)
        new_module.load_state_dict(new_module_state_dict, strict=False)  # bias is not needed.

    def _update_mapped_woqlinear_modules(self, name, format_woqlinear_module, module_kwargs):
        """Checks whether the format mapping module needs to be updated to the device mapping module."""
        # format mapping module class
        OldWeightOnlyLinearClass = format_woqlinear_mapping[self.format]

        # deivice mapping module class
        NewWeightOnlyLinearClass = device_woqlinear_mapping[self.device]

        # if format mapping module doesn't match device mapping module, then replace to device mapping module
        if OldWeightOnlyLinearClass != NewWeightOnlyLinearClass:
            logger.debug(
                f"Replacing {name}'s type from "
                f"'{OldWeightOnlyLinearClass.__name__}' "
                f"to '{NewWeightOnlyLinearClass.__name__}'"
            )

            # initialize the new WeightOnlyLinearClass
            device_mapping_module = NewWeightOnlyLinearClass(**module_kwargs)

            # unpack format mapping module and re-pack to device mapping module
            params_dict = format_woqlinear_module.unpack().to(self.device)
            device_mapping_module.pack(**params_dict)

            # if the new module is HPUWeightOnlyLinear, save hpu format tensor for next loading
            if NewWeightOnlyLinearClass == HPUWeightOnlyLinear and not self._should_save_hpu_format_tensor:
                self._should_save_hpu_format_tensor = True

            return device_mapping_module
        else:
            return format_woqlinear_module

    def _get_model_class_and_config(self):
        from transformers import AutoConfig, AutoModelForCausalLM
        from transformers.dynamic_module_utils import get_class_from_dynamic_module, resolve_trust_remote_code
        from transformers.models.auto.auto_factory import _get_model_class

        # Autofactory
        kwargs_orig = copy.deepcopy(self.kwargs)
        trust_remote_code = self.kwargs.pop("trust_remote_code", None)
        revision = self.kwargs.get("revision", "main")
        kwarg_attn_imp = self.kwargs.pop("attn_implementation", None)

        config = AutoConfig.from_pretrained(
            self.model_name_or_path, trust_remote_code=trust_remote_code, revision=revision
        )
        # quantization_config = config.quantization_config

        if kwarg_attn_imp is not None and config._attn_implementation != kwarg_attn_imp:  # pragma: no cover
            config._attn_implementation = kwarg_attn_imp

        has_remote_code = hasattr(config, "auto_map") and AutoModelForCausalLM.__name__ in config.auto_map

        has_local_code = (
            hasattr(AutoModelForCausalLM, "_model_mapping")
            and type(config) in AutoModelForCausalLM._model_mapping.keys()
        )

        trust_remote_code = resolve_trust_remote_code(
            trust_remote_code,
            self.model_name_or_path,
            has_local_code,
            has_remote_code,
        )

        model_class = self.kwargs.get("model_class", None)
        if model_class:
            return model_class, config

        if has_remote_code and trust_remote_code:  # pragma: no cover
            class_ref = config.auto_map[AutoModelForCausalLM.__name__]
            model_class = get_class_from_dynamic_module(class_ref, self.model_name_or_path, **kwargs_orig)
            if os.path.isdir(self.model_name_or_path):
                model_class.register_for_auto_class(AutoModelForCausalLM.__name__)
            else:
                AutoModelForCausalLM.register(config.__class__, model_class, exist_ok=True)
        elif type(config) in AutoModelForCausalLM._model_mapping.keys():
            model_class = _get_model_class(config, AutoModelForCausalLM._model_mapping)
        else:
            logger.info("Couldn't find model class.")
        return model_class, config

    def _get_loaded_state_dict(self, config):
        import transformers
        from transformers.configuration_utils import PretrainedConfig
        from transformers.modeling_utils import get_checkpoint_shard_files, load_state_dict
        from transformers.utils import cached_file, extract_commit_hash

        if transformers.__version__ < "5.0.0":  # pragma: no cover
            from transformers.utils import is_safetensors_available

        subfolder = self.kwargs.pop("subfolder", "")
        variant = self.kwargs.pop("variant", None)
        cache_dir = self.kwargs.pop("cache_dir", None)
        force_download = self.kwargs.pop("force_download", False)
        proxies = self.kwargs.pop("proxies", None)
        resume_download = self.kwargs.pop("resume_download", False)
        local_files_only = self.kwargs.pop("local_files_only", False)
        offload_folder = self.kwargs.pop("offload_folder", None)
        offload_state_dict = self.kwargs.pop("offload_state_dict", False)
        use_auth_token = self.kwargs.pop("use_auth_token", None)
        token = self.kwargs.pop("token", None)
        from_pipeline = self.kwargs.pop("_from_pipeline", None)
        from_auto_class = self.kwargs.pop("_from_auto", False)
        revision = self.kwargs.pop("revision", "main")
        commit_hash = self.kwargs.pop("_commit_hash", None)
        use_hpu_safetensors = self.kwargs.pop("use_hpu_safetensors", None)
        use_safetensors = self.kwargs.pop("use_safetensors", None)

        if transformers.__version__ < "5.0.0":  # pragma: no cover
            if use_safetensors is None and not is_safetensors_available():
                use_safetensors = False

        if transformers.__version__ < "5.0.0":  # pragma: no cover
            if use_hpu_safetensors is None and not is_safetensors_available():
                use_hpu_safetensors = False
        else:  # pragma: no cover
            if use_hpu_safetensors is None:
                use_hpu_safetensors = False

        if use_auth_token is not None:  # pragma: no cover
            logger.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers. "
                "Please use `token` instead."
            )
            if token is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )
            token = use_auth_token

        user_agent = {
            "file_type": "model",
            "framework": "pytorch",
            "from_auto_class": from_auto_class,
        }
        if from_pipeline is not None:  # pragma: no cover
            user_agent["using_pipeline"] = from_pipeline

        if commit_hash is None:
            if not isinstance(config, PretrainedConfig):  # pragma: no cover
                # We make a call to the config file first (which may be absent)
                # to get the commit hash as soon as possible.
                resolved_config_file = cached_file(
                    self.model_name_or_path,
                    "config.json",
                    cache_dir=cache_dir,
                    force_download=force_download,
                    resume_download=resume_download,
                    proxies=proxies,
                    local_files_only=local_files_only,
                    token=token,
                    revision=revision,
                    subfolder=subfolder,
                    _raise_exceptions_for_missing_entries=False,
                    _raise_exceptions_for_connection_errors=False,
                )
                commit_hash = extract_commit_hash(resolved_config_file, commit_hash)
            else:
                commit_hash = getattr(config, "_commit_hash", None)

        is_sharded = False
        sharded_metadata = None

        self.model_name_or_path = str(self.model_name_or_path)

        # get resolved weight archive file
        kwargs = {
            "use_safetensors": use_safetensors,
            "variant": variant,
            "cache_dir": cache_dir,
            "force_download": force_download,
            "proxies": proxies,
            "resume_download": resume_download,
            "local_files_only": local_files_only,
            "token": token,
            "user_agent": user_agent,
            "revision": revision,
            "subfolder": subfolder,
            "_raise_exceptions_for_gated_repo": False,
            "_raise_exceptions_for_missing_entries": False,
            "_commit_hash": commit_hash,
        }
        resolved_archive_file, is_sharded = self._get_resolved_archive_file(**kwargs)

        self._model_local_dir = os.path.abspath(os.path.expanduser(os.path.dirname(resolved_archive_file)))
        # if hpu format tensor can be used directly, then update resolved_archive_file to the hpu format tensor file
        if self._use_hpu_module():
            resolved_archive_file = os.path.join(self._model_local_dir, HPU_SAFE_WEIGHTS_NAME)
            is_sharded = False

        logger.info(f"Find weight file {resolved_archive_file}")

        if is_sharded:  # pragma: no cover
            # rsolved_archive_file becomes a list of files that point to the different checkpoint shards in this case.
            resolved_archive_file, sharded_metadata = get_checkpoint_shard_files(
                self.model_name_or_path,
                resolved_archive_file,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                local_files_only=local_files_only,
                token=token,
                user_agent=user_agent,
                revision=revision,
                subfolder=subfolder,
                _commit_hash=commit_hash,
            )
            self.kwargs["sharded_metadata"] = sharded_metadata

        # Time to load the checkpoint
        state_dict = None
        if not isinstance(resolved_archive_file, list):
            resolved_archive_file = [resolved_archive_file]
        for shard_file in resolved_archive_file:
            if state_dict is None:
                state_dict = load_state_dict(shard_file)
            else:
                state_dict.update(load_state_dict(shard_file))

        # set kwargs for next functions to use
        self.kwargs["is_sharded"] = is_sharded
        self.kwargs["offload_folder"] = offload_folder
        self.kwargs["offload_state_dict"] = offload_state_dict
        self.kwargs["resolved_archive_file"] = resolved_archive_file

        return state_dict

    def _get_resolved_archive_file(self, **kwargs):
        """Get weight archive file of model."""
        import transformers
        from transformers.modeling_utils import _add_variant

        is_transformers_v5 = transformers.__version__ >= "5.0.0"
        from transformers.utils import (
            SAFE_WEIGHTS_INDEX_NAME,
            SAFE_WEIGHTS_NAME,
            WEIGHTS_INDEX_NAME,
            WEIGHTS_NAME,
            cached_file,
            has_file,
        )

        if not is_transformers_v5:
            from transformers.utils import download_url, is_remote_url

        use_safetensors = kwargs.pop("use_safetensors")
        variant = kwargs.pop("variant")
        subfolder = kwargs.get("subfolder")

        resolved_archive_file = None
        is_sharded = False
        is_local = os.path.isdir(self.model_name_or_path)
        if is_local:  # pragma: no cover
            # self.model_name_or_path is a local directory
            if os.path.isfile(
                os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(WEIGHTS_NAME, variant),
                )
            ):
                # Load from a PyTorch checkpoint
                archive_file = os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(WEIGHTS_NAME, variant),
                )
            elif os.path.isfile(
                os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(WEIGHTS_INDEX_NAME, variant),
                )
            ):
                # Load from a sharded PyTorch checkpoint
                archive_file = os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(WEIGHTS_INDEX_NAME, variant),
                )
                is_sharded = True
            elif os.path.isfile(
                os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(SAFE_WEIGHTS_NAME, variant),
                )
            ):
                # Load from a safetensors checkpoint
                archive_file = os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(SAFE_WEIGHTS_NAME, variant),
                )
            elif os.path.isfile(
                os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(SAFE_WEIGHTS_INDEX_NAME, variant),
                )
            ):
                # Load from a safetensors checkpoint
                archive_file = os.path.join(
                    self.model_name_or_path,
                    subfolder,
                    _add_variant(SAFE_WEIGHTS_INDEX_NAME, variant),
                )
                is_sharded = True
        elif os.path.isfile(os.path.join(subfolder, self.model_name_or_path)):  # pragma: no cover
            archive_file = self.model_name_or_path
            is_local = True
        elif not is_transformers_v5 and is_remote_url(self.model_name_or_path):  # pragma: no cover
            # self.model_name_or_path is a url
            filename = self.model_name_or_path
            resolved_archive_file = download_url(self.model_name_or_path)
        else:
            # self.model_name_or_path is a model_id in huggingface
            if use_safetensors is not False:
                filename = _add_variant(SAFE_WEIGHTS_NAME, variant)
            else:
                filename = _add_variant(WEIGHTS_NAME, variant)
            try:
                # Load from URL or cache if already cached
                cached_file_kwargs = kwargs
                resolved_archive_file = cached_file(self.model_name_or_path, filename, **cached_file_kwargs)

                # Since we set _raise_exceptions_for_missing_entries=False, we don't get an exception but a None
                # result when internet is up, the repo and revision exist, but the file does not.
                if resolved_archive_file is None and filename == _add_variant(
                    SAFE_WEIGHTS_NAME, variant
                ):  # pragma: no cover
                    # Maybe the checkpoint is sharded, we try to grab the index name in this case.
                    resolved_archive_file = cached_file(
                        self.model_name_or_path,
                        _add_variant(SAFE_WEIGHTS_INDEX_NAME, variant),
                        **cached_file_kwargs,
                    )
                    if resolved_archive_file is not None:
                        is_sharded = True
                    elif use_safetensors:
                        raise EnvironmentError(
                            f"{self.model_name_or_path} does not appear to have a file named"
                            f" {_add_variant(SAFE_WEIGHTS_NAME, variant)} or "
                            f"{_add_variant(SAFE_WEIGHTS_INDEX_NAME, variant)} "
                            "and thus cannot be loaded with `safetensors`. Please make sure that the model has "
                            "been saved with `safe_serialization=True` or do not set `use_safetensors=True`."
                        )
                    else:
                        # This repo has no safetensors file of any kind, we switch to PyTorch.
                        filename = _add_variant(WEIGHTS_NAME, variant)
                        resolved_archive_file = cached_file(self.model_name_or_path, filename, **cached_file_kwargs)
                if resolved_archive_file is None and filename == _add_variant(
                    WEIGHTS_NAME, variant
                ):  # pragma: no cover
                    # Maybe the checkpoint is sharded, we try to grab the index name in this case.
                    resolved_archive_file = cached_file(
                        self.model_name_or_path,
                        _add_variant(WEIGHTS_INDEX_NAME, variant),
                        **cached_file_kwargs,
                    )
                    if resolved_archive_file is not None:
                        is_sharded = True

                if resolved_archive_file is None:  # pragma: no cover
                    # Otherwise, maybe there is a TF or Flax model file.  We try those to give a helpful error
                    # message.
                    has_file_kwargs = {
                        "revision": cached_file_kwargs.get("revision"),
                        "proxies": cached_file_kwargs.get("proxies"),
                        "token": cached_file_kwargs.get("token"),
                    }
                    if variant is not None and has_file(self.model_name_or_path, WEIGHTS_NAME, **has_file_kwargs):
                        raise EnvironmentError(
                            f"{self.model_name_or_path} does not appear to have a file named"
                            f" {_add_variant(WEIGHTS_NAME, variant)} but there is a file without the variant"
                            f" {variant}. Use `variant=None` to load this model from those weights."
                        )
                    else:
                        raise EnvironmentError(
                            f"{self.model_name_or_path} does not appear to have a file named"
                            f" {_add_variant(WEIGHTS_NAME, variant)}."
                        )
            except EnvironmentError:  # pragma: no cover
                # Raise any environment error raise by `cached_file`. It will have a helpful error message adapted
                # to the original exception.
                raise
            except Exception as e:  # pragma: no cover
                # For any other exception, we throw a generic error.
                raise EnvironmentError(
                    f"Can't load the model for '{self.model_name_or_path}'. If you were trying to load it"
                    " from 'https://huggingface.co/models', make sure you don't have a local directory with the"
                    f" same name. Otherwise, make sure '{self.model_name_or_path}' is the correct path to a"
                    f" directory containing a file named {_add_variant(WEIGHTS_NAME, variant)}."
                ) from e

        if is_local:
            resolved_archive_file = archive_file

        return resolved_archive_file, is_sharded

    def _init_hf_model(self, model_class, config):
        import transformers
        from accelerate.big_modeling import init_empty_weights

        if transformers.__version__ < "5.0.0":  # pragma: no cover
            from transformers.modeling_utils import no_init_weights
        else:  # pragma: no cover
            from transformers.initialization import no_init_weights
        from transformers.utils import ContextManagers

        _fast_init = self.kwargs.pop("_fast_init", True)
        torch_dtype = self.kwargs.pop("torch_dtype", "auto")
        is_sharded = self.kwargs.pop("is_sharded", False)
        sharded_metadata = self.kwargs.pop("sharded_metadata", None)
        offload_folder = self.kwargs.pop("offload_folder", None)
        offload_state_dict = self.kwargs.pop("offload_state_dict", False)
        resolved_archive_file = self.kwargs.pop("resolved_archive_file", None)

        # set dtype to instantiate the model under:
        # 1. If torch_dtype is not None, we use that dtype
        # 2. If torch_dtype is "auto", we auto-detect dtype from the loaded state_dict,
        #    by checking its first weights entry that is of a floating type
        #    - we assume all floating dtype weights are of the same dtype
        dtype_orig = None
        if torch_dtype is not None:
            if isinstance(torch_dtype, str):
                if torch_dtype == "auto":  # pragma: no cover
                    if (
                        hasattr(config, "torch_dtype")
                        and config.torch_dtype is not None
                        and config.torch_dtype != "auto"
                    ):
                        torch_dtype = config.torch_dtype
                    else:
                        if is_sharded and "dtype" in sharded_metadata:
                            torch_dtype = sharded_metadata["dtype"]
                        else:
                            torch_dtype = torch.float32
                else:  # pragma: no cover
                    assert False, f'`torch_dtype` can be either `torch.dtype` or `"auto"`, but received {torch_dtype}'

            if parse(transformers.__version__) >= parse("5.0.0"):
                from transformers.modeling_utils import local_torch_dtype

                local_torch_dtype(torch_dtype, model_class.__name__)
                dtype_orig = None
            elif parse(transformers.__version__) >= parse("4.56.0"):
                dtype_orig = model_class._set_default_dtype(torch_dtype)
            else:
                dtype_orig = model_class._set_default_torch_dtype(torch_dtype)

        init_contexts = (
            [no_init_weights(_enable=_fast_init)]
            if parse(transformers.__version__) < parse("4.51")
            else [no_init_weights()]
        )
        init_contexts.append(init_empty_weights())

        with ContextManagers(init_contexts):
            model = model_class(config, **self.kwargs)

        # set kwargs for next functions to use
        self.kwargs["resolved_archive_file"] = resolved_archive_file
        self.kwargs["torch_dtype"] = torch_dtype
        self.kwargs["dtype_orig"] = dtype_orig
        self.kwargs["offload_folder"] = offload_folder
        self.kwargs["offload_state_dict"] = offload_state_dict

        return model

    def _load_remaining_pretrained_weight(self, model):
        """Load remaining pretrained weight.

        In _build_woq_model function, linear will be replaced to weight-only quantization linear
        and its quantized weight will be loaded. Remaining pretrained weight (like layernorm weight,
        embedding weight or other unquantized linear weight) will be loaded in this function.
        """
        from transformers.modeling_utils import load_state_dict

        resolved_archive_file = self.kwargs.pop("resolved_archive_file", None)
        torch_dtype = self.kwargs.pop("torch_dtype", torch.float32)
        dtype_orig = self.kwargs.pop("dtype_orig", None)

        # restore default dtype
        if dtype_orig is not None:
            torch.set_default_dtype(dtype_orig)

        if not isinstance(resolved_archive_file, list):
            resolved_archive_file = [resolved_archive_file]
        for shard_file in resolved_archive_file:
            state_dict = load_state_dict(shard_file)
            keys = list(state_dict.keys())
            for k in keys:
                if k not in self.loaded_state_dict_keys:
                    state_dict.pop(k)
            for k, v in state_dict.items():
                state_dict[k] = v.to(torch_dtype)
            model.load_state_dict(state_dict, strict=False, assign=True)

        # make sure token embedding weights are still tied if needed
        model.tie_weights()

        # Set model in evaluation mode to deactivate DropOut modules by default
        model.eval()

        return model

    def _save_hpu_format_tensor(self, model):  # pragma: no cover
        from safetensors.torch import save_file

        if not os.path.exists(self._model_local_dir):
            logger.warning(f"{self._model_local_dir} doesn't exist, can't save hpu format safetensors")

        if self.format == SaveLoadFormat.HUGGINGFACE:
            filename = os.path.join(self._model_local_dir, HPU_SAFE_WEIGHTS_NAME)
            # remove tied weights
            state_dict = model.state_dict()
            if getattr(getattr(model, "config", None), "tie_word_embeddings", False):
                for key in model._tied_weights_keys:
                    state_dict.pop(key)
            save_file(state_dict, filename=filename, metadata={"format": "pt"})
            logger.debug(f"Save hpu format tensor to {filename}")
        elif self.format == SaveLoadFormat.DEFAULT:
            if self.sharded_checkpoints:
                qmodel_weight_file_path = os.path.join(self._model_local_dir, SHARDED_HPU_WEIGHT_NAME)
                save_file(model.state_dict(), qmodel_weight_file_path)
            else:
                qmodel_weight_file_path = os.path.join(self._model_local_dir, HPU_WEIGHT_NAME)
                torch.save(model.state_dict(), qmodel_weight_file_path)
            logger.debug(f"Save hpu format tensor to {qmodel_weight_file_path}")

    def _use_hpu_module(self):  # pragma: no cover
        """Check whether hpu weight-only quantization linear module can be used.

        return True when:
        1. device is 'hpu'
        2. model has hpu format tensor in local cache directory:
            - has 'hpu_model.safetensors' file with huggingface format
            - or has 'quantized_hpu_weight.pt' file with default format
           or 'format' flag in config.json file is 'habana' (flag name needs discussion, not implemented yet)
        """
        if self.device == "hpu" and os.path.exists(self._model_local_dir):
            if self.format == SaveLoadFormat.HUGGINGFACE:
                if os.path.exists(os.path.join(self._model_local_dir, HPU_SAFE_WEIGHTS_NAME)):
                    # update resolved_archive_file
                    self.kwargs["resolved_archive_file"] = os.path.join(self._model_local_dir, HPU_SAFE_WEIGHTS_NAME)
                    return True
            elif self.format == SaveLoadFormat.DEFAULT:
                if self.sharded_checkpoints:
                    if os.path.exists(os.path.join(self._model_local_dir, SHARDED_HPU_WEIGHT_NAME)):
                        return True
                else:
                    if os.path.exists(os.path.join(self._model_local_dir, HPU_WEIGHT_NAME)):
                        return True
        return False


def change_config_to_hf_format(config_mappings):
    """Change INC config_mappings to Huggingface format."""
    # Refer to https://huggingface.co/TheBloke/Llama-2-7B-Chat-GPTQ/blob/main/config.json
    default_quantization_config = {
        "bits": 4,
        "group_size": 128,
        "damp_percent": 0.01,
        "desc_act": True,
        "sym": True,
        "true_sequential": True,
        "model_name_or_path": None,
        "model_file_base_name": "model",
        "quant_method": "gptq",  # INC is using AutoGPTQ format for RTN, GPTQ, AWQ, and TEQ
    }

    def _is_lm_head(name):
        for lm_head_name in LM_HEAD_NAMES:
            if re.match(lm_head_name, name):
                if config.dtype != "fp32":
                    raise ValueError(f"{name} should not be quantized if you want to save in huggingface format.")
                else:
                    return True
        return False

    bits, group_size, damp_percent, desc_act, sym, true_sequential = None, None, None, None, None, None
    for (name, type), config in config_mappings.items():
        # Check 1: check whether LM_HEAD_NAMES is quantized, hf_format cannot quantize LM_HEAD_NAMES
        if _is_lm_head(name):
            continue
        # Check 2: check whether config is the same for all modules
        if bits is None:
            # percdamp, act_order, true_sequential are specific to GPTQ
            bits = config.bits
            group_size = config.group_size
            sym = config.use_sym
            damp_percent = config.percdamp if hasattr(config, "percdamp") else 0
            desc_act = config.act_order if hasattr(config, "act_order") else False
            true_sequential = config.true_sequential if hasattr(config, "true_sequential") else False
        else:
            assert bits == config.bits, "bits should be the same for all modules, got {bits} and {config.bits}."
            assert sym == config.use_sym, "sym should be the same for all modules, got {sym} and {config.use_sym}."
            assert (
                group_size == config.group_size
            ), "group_size should be the same for all modules, got {group_size} and {config.group_size}."
            if hasattr(config, "percdamp"):
                assert (
                    damp_percent == config.percdamp
                ), "percdamp should be the same for all modules, got {damp_percent} and {config.percdamp}."
            if hasattr(config, "act_order"):
                assert (
                    desc_act == config.act_order
                ), "act_order should be the same for all modules, got {desc_act} and {config.act_order}."
            if hasattr(config, "true_sequential"):
                assert (
                    true_sequential == config.true_sequential
                ), "true_sequential should be the same for all modules, got {true_sequential} and {config.true_sequential}."
    default_quantization_config["bits"] = bits
    default_quantization_config["group_size"] = group_size
    default_quantization_config["damp_percent"] = damp_percent
    default_quantization_config["desc_act"] = desc_act
    default_quantization_config["sym"] = sym
    default_quantization_config["true_sequential"] = true_sequential
    return default_quantization_config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
from datetime import timedelta
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

import torch
import torch.nn.functional as F
import transformers
from accelerate import Accelerator, DistributedType, InitProcessGroupKwargs, find_executable_batch_size
from lm_eval import utils
from lm_eval.api.instance import Instance
from lm_eval.api.model import TemplateLM
from lm_eval.models.utils import Collator, clear_torch_cache, get_dtype, pad_and_concat, stop_sequences_criteria
from packaging import version
from packaging.version import Version
from peft import PeftModel
from peft import __version__ as PEFT_VERSION
from tqdm import tqdm
from transformers import AutoModel, AutoModelForCausalLM, AutoModelForSeq2SeqLM
from transformers.models.auto.modeling_auto import (
    MODEL_FOR_CAUSAL_LM_MAPPING_NAMES,
    MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING_NAMES,
)

try:
    from lm_eval.evaluator_utils import eval_logger
except ImportError:
    from lm_eval.utils import eval_logger

# transformers.AutoModelForCausalLM = AutoModelForCausalLM
# transformers.AutoModelForSeq2SeqLM = AutoModelForSeq2SeqLM
# transformers.AutoModel = AutoModel


def _get_accelerate_args(
    device_map_option: Optional[str] = "auto",
    max_memory_per_gpu: Optional[Union[int, str]] = None,
    max_cpu_memory: Optional[Union[int, str]] = None,
    offload_folder: Optional[str] = "./offload",
) -> dict:
    """Returns the kwargs needed to apply `accelerate` in `AutoModel.from_pretrained`."""
    max_memory = {}
    if max_memory_per_gpu is not None:
        max_memory_per_gpu_map = {device_idx: max_memory_per_gpu for device_idx in range(torch.cuda.device_count())}
        max_memory.update(max_memory_per_gpu_map)
    if max_cpu_memory is not None:
        max_memory["cpu"] = max_cpu_memory

    args = {}
    if max_memory:
        args["max_memory"] = max_memory
    args["device_map"] = device_map_option
    args["offload_folder"] = offload_folder
    return args


class HFLM(TemplateLM):
    """An abstracted Huggingface model class. Enables usage with both models of
    `neural_compressor.transformers.AutoModelForCausalLM` and
    `neural_compressor.transformers.AutoModelForSeq2SeqLM` classes.

    Supports data-parallel multi-GPU with HF Accelerate.
    """

    AUTO_MODEL_CLASS = None
    _DEFAULT_MAX_LENGTH = 2048

    def __init__(
        self,
        pretrained: Optional[Union[str, transformers.PreTrainedModel]] = "gpt2",
        backend: Optional[Literal["default", "causal", "seq2seq"]] = "default",
        # override whether the model should be treated as decoder-only (causal) or encoder-decoder (seq2seq)
        revision: Optional[str] = "main",
        subfolder: Optional[str] = None,
        tokenizer: Optional[
            Union[
                str,
                transformers.PreTrainedTokenizer,
                transformers.PreTrainedTokenizerFast,
            ]
        ] = None,
        truncation: Optional[bool] = False,
        logits_cache: bool = True,
        max_length: Optional[int] = None,
        device: Optional[str] = "cuda",
        dtype: Optional[Union[str, torch.dtype]] = "auto",
        batch_size: Optional[Union[int, str]] = 1,
        max_batch_size: Optional[int] = 64,
        trust_remote_code: Optional[bool] = False,
        use_fast_tokenizer: Optional[bool] = True,
        add_bos_token: Optional[bool] = False,
        # arguments used for splitting a model across GPUs naively.
        # only used if `parallelize=True`.
        parallelize: Optional[bool] = False,
        device_map_option: Optional[str] = "auto",
        max_memory_per_gpu: Optional[Union[int, str]] = None,
        max_cpu_memory: Optional[Union[int, str]] = None,
        offload_folder: Optional[Union[str, os.PathLike]] = "./offload",
        # PEFT and quantization options
        peft: Optional[str] = None,
        autogptq: Optional[Union[bool, str]] = False,
        pad_to_buckets: Optional[Union[bool]] = False,
        buckets: Optional[list] = [32, 64, 128, 256, 512, 1024, 2048, 4096],
        model_format: Optional[str] = "torch",
        **kwargs,
    ) -> None:
        super().__init__()
        self.pad_to_buckets = pad_to_buckets
        self.buckets = buckets
        self.last_bucket = -1
        self.model_format = model_format
        # optionally: take in an already-initialized transformers.PreTrainedModel
        if not isinstance(pretrained, str):
            eval_logger.warning(
                "`pretrained` model kwarg is not of type `str`. "
                + "Many other model arguments may be ignored. "
                + "Please do not launch via accelerate or use `parallelize=True`"
                + "if passing an existing model this way."
            )
            assert (
                not parallelize
            ), "`parallelize=True` is not compatible with passing pre-initialized model to `pretrained`"
            self._model = pretrained
            self._device = self._model.device
            self._config = self._model.config
            gpus = 0

            if tokenizer:
                assert isinstance(tokenizer, transformers.PreTrainedTokenizer) or isinstance(
                    tokenizer, transformers.PreTrainedTokenizerFast
                )
                self.tokenizer = tokenizer
            else:
                # Get tokenizer
                model_name = self._model.name_or_path
                self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                    model_name,
                    revision=revision,
                    trust_remote_code=trust_remote_code,
                    use_fast=use_fast_tokenizer,
                )

        else:
            assert isinstance(device, str)
            assert isinstance(pretrained, str)
            assert isinstance(batch_size, (int, str))

            gpus = torch.cuda.device_count()
            accelerator_kwargs = InitProcessGroupKwargs(timeout=timedelta(weeks=52))
            accelerator = Accelerator(kwargs_handlers=[accelerator_kwargs])
            if accelerator.num_processes > 1:
                self.accelerator = accelerator

            if not (parallelize or accelerator.num_processes > 1):
                # use user-passed device
                device_list = set(
                    ["cuda", "cpu", "xpu", "hpu"]
                    + [f"cuda:{i}" for i in range(torch.cuda.device_count())]
                    + ["mps", "mps:0"]
                )
                if device and device in device_list:
                    self._device = torch.device(device)
                    eval_logger.info(f"Using device '{device}'")
                    if device in ("mps", "mps:0") and version.parse(torch.__version__) < version.parse("2.1"):
                        raise RuntimeError(f"mps requires torch >= 2.1. You have {torch.__version__}")
                else:
                    eval_logger.info("Device not specified")
                    eval_logger.info(f"Cuda Available? {torch.cuda.is_available()}")
                    self._device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
            else:
                if device != "cuda":
                    eval_logger.info(
                        f"Using `accelerate launch` or `parallelize=True`, device '{device}' "
                        + "will be overridden when placing model."
                    )
                # TODO: include in warning that `load_in_8bit` etc. affect this too
                self._device = torch.device(device)

            # TODO: update this to be less of a hack once subfolder is fixed in HF
            revision = revision + ("/" + subfolder if subfolder is not None else "")

            self._get_config(
                pretrained,
                revision=revision,
                trust_remote_code=trust_remote_code,
            )

        # determine which of 'causal' and 'seq2seq' backends to use
        self._get_backend(config=self.config, backend=backend, trust_remote_code=trust_remote_code)

        # if we passed `pretrained` as a string, initialize our model now
        if isinstance(pretrained, str):
            self._create_model(
                pretrained=pretrained,
                revision=revision,
                dtype=dtype,
                trust_remote_code=trust_remote_code,
                parallelize=parallelize,
                device_map_option=device_map_option,
                max_memory_per_gpu=max_memory_per_gpu,
                max_cpu_memory=max_cpu_memory,
                offload_folder=offload_folder,
                peft=peft,
                autogptq=autogptq,
                **kwargs,
            )

        # access self._model through self.model property outside this method
        if isinstance(self.model, torch.nn.Module):
            self.model.eval()
            self.model.tie_weights()

        if isinstance(pretrained, str) and (gpus >= 1 or str(self.device) == "mps"):
            # TODO: can remove this whole snippet except in the mps case, perhaps?
            if not (parallelize or autogptq or hasattr(self, "accelerator")):
                # place model onto device requested manually,
                # if not using HF Accelerate or device_map
                # or any other option that preloads model onto device
                try:
                    self.model.to(self.device)
                except ValueError:
                    eval_logger.debug(
                        "Failed to place model onto specified device. "
                        + "This may be because the model is quantized via `bitsandbytes` or `device_map` is provided."
                        + "If the desired GPU is being used, this message is safe to ignore."
                    )

        self._create_tokenizer(
            pretrained,
            tokenizer,
            revision=revision,
            trust_remote_code=trust_remote_code,
            use_fast_tokenizer=use_fast_tokenizer,
        )

        self.truncation = truncation
        self.logits_cache = logits_cache
        self.vocab_size = self.tokenizer.vocab_size
        # select (or create) a pad token to use
        if self.tokenizer.pad_token:
            pass
        elif self.tokenizer.unk_token:
            self.tokenizer.pad_token_id = self.tokenizer.unk_token_id
        elif self.tokenizer.eos_token:
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id
        else:
            if getattr(self.config, "model_type", None) == "qwen":
                # Qwen's trust_remote_code tokenizer does not allow for adding special tokens
                self.tokenizer.pad_token = "<|endoftext|>"
            elif (
                self.tokenizer.__class__.__name__ == "RWKVWorldTokenizer"
                or self.tokenizer.__class__.__name__ == "Rwkv5Tokenizer"
            ):
                # The RWKV world tokenizer, does not allow for adding special tokens /
                # setting the pad token (which is set as 0)
                # The additional tokenizer name check is needed, as there exists rwkv4 models with neox tokenizer
                # ---
                # Note that the world tokenizer class name, might change in the future
                # for the final huggingface merge
                # https://github.com/huggingface/transformers/pull/26963
                assert self.tokenizer.pad_token_id == 0
            else:
                self.tokenizer.add_special_tokens({"pad_token": "<|pad|>"})

        # TODO: override this for Gemma
        self.add_bos_token = add_bos_token
        if getattr(self.config, "model_type", None) == "gemma":
            self.add_bos_token = True
            eval_logger.info(
                f"Model type is '{self.config.model_type}', "
                + "a BOS token will be used as Gemma underperforms without it."
            )

        self._max_length = max_length

        self.batch_schedule = 1
        self.batch_sizes = {}
        self.max_batch_size = max_batch_size

        if str(batch_size).startswith("auto"):
            batch_size = batch_size.split(":")
            self.batch_size_per_gpu = batch_size[0]
            self.batch_schedule = float(batch_size[1]) if len(batch_size) > 1 else 1
        else:
            self.batch_size_per_gpu = int(batch_size)

        if isinstance(pretrained, str):
            # multigpu data-parallel support when launched with accelerate
            if gpus > 1:
                if parallelize:
                    if accelerator.num_processes > 1:
                        raise RuntimeError(
                            "Attempted to use both a HF Accelerate `device_map` "
                            + "and to launch via `accelerate launch`."
                            + "If this is the case, please either remove `parallelize=True` from"
                            + " --model_args or launch outside of the Accelerate launcher."
                        )
                    else:
                        pass
                elif accelerator.num_processes == 1:
                    # if we aren't launching via accelerate, ditch
                    self._rank = 0
                    self._world_size = 1
                else:
                    if gpus > accelerator.num_processes:
                        eval_logger.warning(
                            "WARNING: The number of total system GPUs "
                            + "does not match the number of spawned processes. "
                            "If you would like to use data parallelism, please launch the script "
                            "with 'accelerate launch *script*'. "
                            f"Current run will proceed with {accelerator.num_processes} devices."
                        )
                    assert accelerator.distributed_type in [
                        DistributedType.FSDP,
                        DistributedType.MULTI_GPU,
                    ], "Unsupported distributed type provided. Only DDP and FSDP are supported."
                    if accelerator.distributed_type == DistributedType.FSDP:
                        self._model = accelerator.prepare(self.model)
                    else:
                        self._model = accelerator.prepare_model(self.model, evaluation_mode=True)
                    self._device = torch.device(f"cuda:{accelerator.local_process_index}")
                    self.accelerator = accelerator

                    if self.accelerator.is_local_main_process:
                        eval_logger.info(f"Using {gpus} devices with data parallelism")

                    self._rank = self.accelerator.local_process_index
                    self._world_size = self.accelerator.num_processes
        else:
            # if a PreTrainedModel was passed into HFLM, we forgo distributed setup.
            eval_logger.warning(
                "Passed an already-initialized model through `pretrained`,"
                + " assuming single-process call to evaluate() or custom distributed integration"
            )
            self._rank = 0
            self._world_size = 1

    @property
    def config(self):
        # return the associated transformers.AutoConfig for the given pretrained model.
        return self._config

    @property
    def model(self):
        # returns the model, unwrapping it if using Accelerate
        if hasattr(self, "accelerator") and self._model is not None:
            return self.accelerator.unwrap_model(self._model)
        else:
            return self._model

    @property
    def eot_token_id(self):
        # we use EOT because end of *text* is more accurate for what we're doing than end of *sentence*
        return self.tokenizer.eos_token_id

    @property
    def max_length(self):
        if self._max_length:  # if max length manually set, return it
            return self._max_length
        seqlen_config_attrs = ("n_positions", "max_position_embeddings", "n_ctx")
        for attr in seqlen_config_attrs:
            if hasattr(self.model.config, attr):
                return getattr(self.model.config, attr)
        if hasattr(self.tokenizer, "model_max_length"):
            if self.tokenizer.model_max_length == 1000000000000000019884624838656:
                return self._DEFAULT_MAX_LENGTH
            return self.tokenizer.model_max_length
        return self._DEFAULT_MAX_LENGTH

    @property
    def max_gen_toks(self) -> int:
        return 256

    @property
    def batch_size(self):
        return self.batch_size_per_gpu

    @property
    def device(self):
        return self._device

    @property
    def rank(self):
        return self._rank

    @property
    def world_size(self):
        return self._world_size

    def _get_backend(
        self,
        config: Union[transformers.PretrainedConfig, transformers.AutoConfig],
        backend: Optional[Literal["default", "causal", "seq2seq"]] = "default",
        trust_remote_code: Optional[bool] = False,
    ) -> None:
        """Helper method during initialization.

        Determines the backend ("causal" (decoder-only) or "seq2seq" (encoder-decoder))
        model type to be used.
        """
        assert backend in ["default", "causal", "seq2seq"]
        if backend != "default":
            # if we've settled on non-default backend, use that manually
            if backend == "causal":
                self.AUTO_MODEL_CLASS = AutoModelForCausalLM
            elif backend == "seq2seq":
                self.AUTO_MODEL_CLASS = AutoModelForSeq2SeqLM
            eval_logger.info(f"Overrode HF model backend type, and using type '{backend}'")
        else:
            # determine and use the default HF backend for this model, based on its config + metadata.
            if getattr(config, "model_type") in MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING_NAMES:
                # first check if model type is listed under seq2seq models, since some
                # models like MBart are listed in both seq2seq and causal mistakenly in HF transformers.
                # these special cases should be treated as seq2seq models.
                self.AUTO_MODEL_CLASS = AutoModelForSeq2SeqLM
            elif getattr(self.config, "model_type") in MODEL_FOR_CAUSAL_LM_MAPPING_NAMES:
                self.AUTO_MODEL_CLASS = AutoModelForCausalLM
            else:
                if not trust_remote_code:
                    eval_logger.warning(
                        "HF model type is neither marked as CausalLM or Seq2SeqLM. \
                    This is expected if your model requires `trust_remote_code=True` but may be an error otherwise."
                    )
                # if model type is neither in HF transformers causal or seq2seq model registries
                # then we default to AutoModelForCausalLM
                self.AUTO_MODEL_CLASS = AutoModelForCausalLM

        assert self.AUTO_MODEL_CLASS in [
            AutoModelForCausalLM,
            AutoModelForSeq2SeqLM,
        ]
        return None

    def _get_config(
        self,
        pretrained: str,
        revision: str = "main",
        trust_remote_code: bool = False,
    ) -> None:
        self._config = transformers.AutoConfig.from_pretrained(
            pretrained,
            revision=revision,
            trust_remote_code=trust_remote_code,
        )

    def _create_model(
        self,
        pretrained: str,
        revision: Optional[str] = "main",
        dtype: Optional[Union[str, torch.dtype]] = "auto",
        trust_remote_code: Optional[bool] = False,
        # arguments used for splitting a model across GPUs naively.
        # only used if `parallelize=True`.
        # (accelerate naive PP (device_map) options)
        parallelize: Optional[bool] = False,
        device_map_option: Optional[str] = "auto",
        max_memory_per_gpu: Optional[Union[int, str]] = None,
        max_cpu_memory: Optional[Union[int, str]] = None,
        offload_folder: Optional[str] = "./offload",
        # PEFT and quantization options
        peft: Optional[str] = None,
        autogptq: Optional[Union[bool, str]] = False,
        **kwargs,
    ) -> None:
        """Initializes an HF or HF-compatible PreTrainedModel from scratch
        inside HFLM, using the kwargs passed into self.__init__().

        Also handles functionality such as AutoGPTQ usage and PEFT wrapping.

        For future similar extensions to AutoGPTQ that are not core to HF's ecosystem,
        (such as PyTorch models that are nearly, but not quite, fully mirroring
        HF's public interface relied on in this HFLM class)
        please consider subclassing HFLM and overriding this and other methods as needed.
        """

        model_kwargs = kwargs if kwargs else {}

        if parallelize:
            model_kwargs.update(
                _get_accelerate_args(
                    device_map_option,  # TODO: phase out device_map_option?
                    max_memory_per_gpu,
                    max_cpu_memory,
                    offload_folder,
                )
            )
        elif "device_map" not in model_kwargs:
            # set a device_map to initialize model on the right GPU.
            # this is needed because it seems that the default behavior
            # for quantized models now seems to be device_map="auto"
            # which breaks data-parallel mode.
            if hasattr(self, "accelerator"):
                model_kwargs.update({"device_map": {"": f"cuda:{self.accelerator.local_process_index}"}})
            else:
                model_kwargs.update({"device_map": {"": str(self.device)}})

        if not autogptq:
            if model_kwargs.get("load_in_4bit", None):
                assert transformers.__version__ >= "4.30.0", "load_in_4bit requires transformers >= 4.30.0"
            if transformers.__version__ >= "4.30.0":
                if model_kwargs.get("load_in_4bit", None):
                    if model_kwargs.get("bnb_4bit_compute_dtype", None):
                        model_kwargs["bnb_4bit_compute_dtype"] = get_dtype(model_kwargs["bnb_4bit_compute_dtype"])

            if self.model_format == "onnx" and self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
                if (
                    not os.path.exists(os.path.join(pretrained, "decoder_model.onnx"))
                    and not os.path.exists(os.path.join(pretrained, "decoder_with_past_model.onnx"))
                    and not os.path.exists(os.path.join(pretrained, "decoder_model_merged.onnx"))
                    and not os.path.exists(os.path.join(pretrained, "model.onnx"))
                ):
                    raise ValueError(
                        "Couldn't find any ONNX model name in "
                        + "['decoder_model.onnx', 'decoder_with_past_model.onnx', "
                        "'decoder_model_merged.onnx', 'model.onnx'] in {}.".format(pretrained)
                    )

                import onnxruntime as ort
                import optimum.version
                from optimum.onnxruntime import ORTModelForCausalLM
                from transformers import PretrainedConfig

                model_config = PretrainedConfig.from_pretrained(pretrained)
                sess_options = ort.SessionOptions()
                sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL

                if Version(optimum.version.__version__) >= Version("1.14.0"):
                    if os.path.exists(os.path.join(pretrained, "model.onnx")):
                        session = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                            os.path.join(pretrained, "model.onnx"), session_options=sess_options
                        )
                        inputs_names = [input.name for input in session.get_inputs()]  # pylint: disable=E1101
                        key_value_input_names = [key for key in inputs_names if (".key" in key) or (".value" in key)]
                        use_cache = len(key_value_input_names) > 0

                        self._model = ORTModelForCausalLM(
                            session,  # pylint: disable=E1120
                            model_config,
                            use_cache=True if use_cache else False,
                            use_io_binding=True if use_cache else False,
                        )
                    else:
                        if os.path.exists(os.path.join(pretrained, "decoder_model_merged.onnx")):
                            session = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                                os.path.join(pretrained, "decoder_model_merged.onnx"), session_options=sess_options
                            )
                            self._model = ORTModelForCausalLM(
                                session, model_config, use_cache=True  # pylint: disable=E1120
                            )
                        elif os.path.exists(os.path.join(pretrained, "decoder_with_past_model.onnx")):
                            session = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                                os.path.join(pretrained, "decoder_with_past_model.onnx"), session_options=sess_options
                            )
                            self._model = ORTModelForCausalLM(
                                session, model_config, use_cache=True  # pylint: disable=E1120
                            )
                        elif os.path.exists(os.path.join(pretrained, "decoder_model.onnx")):
                            session = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                                os.path.join(pretrained, "decoder_model.onnx"), session_options=sess_options
                            )
                            self._model = ORTModelForCausalLM(
                                session, model_config, use_cache=False, use_io_binding=False  # pylint: disable=E1120
                            )
                else:
                    if os.path.exists(os.path.join(pretrained, "model.onnx")):
                        session = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                            os.path.join(pretrained, "model.onnx"), session_options=sess_options
                        )
                        inputs_names = session.get_inputs()
                        key_value_input_names = [key for key in inputs_names if (".key" in key) or (".value" in key)]
                        use_cache = len(key_value_input_names) > 0

                        # pylint: disable=E1121,E1124
                        self._model = ORTModelForCausalLM(
                            session[0],
                            model_config,
                            pretrained,
                            use_cache=True if use_cache else False,
                            use_io_binding=True if use_cache else False,
                        )
                    else:
                        if os.path.exists(os.path.join(pretrained, "decoder_model_merged.onnx")):
                            sessions = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                                os.path.join(pretrained, "decoder_model_merged.onnx"), session_options=sess_options
                            )
                            self._model = ORTModelForCausalLM(
                                sessions[0], model_config, pretrained, use_cache=True  # pylint: disable=E1121
                            )
                        elif os.path.exists(os.path.join(pretrained, "decoder_with_past_model.onnx")):
                            sessions = ORTModelForCausalLM.load_model(  # pylint: disable=E1123
                                os.path.join(pretrained, "decoder_model.onnx"),
                                os.path.join(pretrained, "decoder_with_past_model.onnx"),
                                session_options=sess_options,
                            )
                            self._model = ORTModelForCausalLM(
                                sessions[0],  # pylint: disable=E1121
                                model_config,
                                pretrained,
                                sessions[1],
                                use_cache=True,
                            )
                        else:
                            # pylint: disable=E1123,E1124
                            sessions = ORTModelForCausalLM.load_model(
                                os.path.join(pretrained, "decoder_model.onnx"), session_options=sess_options
                            )
                            # pylint: disable=E1121,E1124
                            self._model = ORTModelForCausalLM(
                                sessions[0], model_config, pretrained, use_cache=False, use_io_binding=False
                            )
            elif self.model_format == "onnx" and self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
                if not os.path.exists(os.path.join(pretrained, "encoder_model.onnx")) or (
                    not os.path.exists(os.path.join(pretrained, "decoder_model.onnx"))
                    and not os.path.exists(os.path.join(pretrained, "decoder_model_merged.onnx"))
                ):
                    raise ValueError(
                        "Please ensure encoder_model.onnx and "
                        "decoder_model(_merged).onnx are under {}.".format(pretrained)
                    )

                import onnxruntime as ort
                from optimum.onnxruntime import ORTModelForSeq2SeqLM
                from transformers import PretrainedConfig

                model_config = PretrainedConfig.from_pretrained(pretrained)
                sess_options = ort.SessionOptions()
                sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
                if os.path.exists(os.path.join(pretrained, "decoder_model_merged.onnx")):
                    sessions = ORTModelForSeq2SeqLM.load_model(
                        os.path.join(pretrained, "encoder_model.onnx"),
                        os.path.join(pretrained, "decoder_model_merged.onnx"),
                    )

                    self._model = ORTModelForSeq2SeqLM(
                        sessions[0],
                        sessions[1],
                        model_config,
                        pretrained,
                        use_cache=True,
                    )

                elif os.path.exists(os.path.join(pretrained, "decoder_with_past_model.onnx")):
                    sessions = ORTModelForSeq2SeqLM.load_model(
                        os.path.join(pretrained, "encoder_model.onnx"),
                        os.path.join(pretrained, "decoder_model.onnx"),
                        os.path.join(pretrained, "decoder_with_past_model.onnx"),
                    )

                    self._model = ORTModelForSeq2SeqLM(
                        sessions[0],
                        sessions[1],
                        model_config,
                        pretrained,
                        sessions[2],
                        use_cache=True,
                    )
                else:
                    sessions = ORTModelForSeq2SeqLM.load_model(  # pylint: disable=E1120
                        os.path.join(pretrained, "encoder_model.onnx"),
                        os.path.join(pretrained, "decoder_model.onnx"),
                    )

                    self._model = ORTModelForSeq2SeqLM(
                        sessions[0],
                        sessions[1],
                        model_config,
                        pretrained,
                        use_cache=False,
                        use_io_binding=False,
                    )
            else:
                dtype = "float32" if dtype == "auto" else dtype
                if pretrained == "Muennighoff/tiny-random-bert":
                    self._model = None
                else:
                    self._model = self.AUTO_MODEL_CLASS.from_pretrained(
                        pretrained,
                        revision=revision,
                        torch_dtype=get_dtype(dtype),
                        trust_remote_code=trust_remote_code,
                        **model_kwargs,
                    )
        else:
            try:
                from auto_gptq import AutoGPTQForCausalLM
            except ModuleNotFoundError:
                raise Exception(
                    "Tried to load auto_gptq, but auto-gptq is not installed ",
                    "please install auto-gptq via pip install lm-eval[gptq] or pip install -e .[gptq]",
                )

            self._model = AutoGPTQForCausalLM.from_quantized(
                pretrained,
                trust_remote_code=trust_remote_code,
                model_basename=None if autogptq is True else Path(autogptq).stem,
                use_safetensors=(True if autogptq is True else autogptq.endswith(".safetensors")),
                **model_kwargs,
            )

        if peft:
            if model_kwargs.get("load_in_4bit", None):
                assert PEFT_VERSION >= "0.4.0", "load_in_4bit requires peft >= 0.4.0"
            self._model = PeftModel.from_pretrained(self._model, peft, revision=revision)

        return None

    def _create_tokenizer(
        self,
        pretrained: Union[str, transformers.PreTrainedModel],
        tokenizer: Optional[
            Union[
                str,
                transformers.PreTrainedTokenizer,
                transformers.PreTrainedTokenizerFast,
            ]
        ],
        revision: Optional[str] = "main",
        trust_remote_code: Optional[bool] = False,
        use_fast_tokenizer: Optional[bool] = True,
    ) -> None:
        """Helper method during initialization.

        Create a tokenizer object corresponding to the correct
        tokenizer for value of `pretrained`, or use the pre-initialized tokenizer passed.
        """

        if tokenizer:
            if isinstance(tokenizer, str):
                self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                    tokenizer,
                    revision=revision,
                    trust_remote_code=trust_remote_code,
                    use_fast=use_fast_tokenizer,
                )
            else:
                assert isinstance(tokenizer, transformers.PreTrainedTokenizer) or isinstance(
                    tokenizer, transformers.PreTrainedTokenizerFast
                )
                self.tokenizer = tokenizer
        else:
            # Get tokenizer based on 'pretrained'
            if isinstance(pretrained, str):
                model_name = pretrained
            else:
                # get the HF hub name via accessor on model
                model_name = self.model.name_or_path

            # chatglm2 tokenizer doesn't support loading from local.
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "auto_map")
                and "chatglm2" in self.model.config.auto_map["AutoConfig"]
            ):
                self.tokenizer = transformers.AutoTokenizer.from_pretrained("THUDM/chatglm2-6b", trust_remote_code=True)
            else:
                self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                    model_name,
                    revision=revision,
                    trust_remote_code=trust_remote_code,
                    use_fast=use_fast_tokenizer,
                )
        return None

    def _detect_batch_size(self, requests=None, pos: int = 0):
        if requests:
            _, context_enc, continuation_enc = requests[pos]
            max_length = len((context_enc + continuation_enc)[-(self.max_length + 1) :][:-1])
            max_context_enc = len(context_enc[-(self.max_length + 1) :])
            max_cont_enc = len(continuation_enc[-(self.max_length + 1) :])
        else:
            max_length = self.max_length

        # if OOM, then halves batch_size and tries again
        @find_executable_batch_size(starting_batch_size=self.max_batch_size)
        def forward_batch(batch_size):
            if self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
                length = max(max_context_enc, max_cont_enc)
                batched_conts = torch.ones((batch_size, length), device=self.device).long()
                test_batch = torch.ones((batch_size, length), device=self.device).long()
                call_kwargs = {
                    "attn_mask": test_batch,
                    "labels": batched_conts,
                }
            else:
                call_kwargs = {}
                test_batch = torch.ones((batch_size, max_length), device=self.device).long()
            for _ in range(5):
                out = F.log_softmax(self._model_call(test_batch, **call_kwargs), dim=-1)

            return batch_size

        try:
            batch_size = forward_batch()  # pylint: disable=E1120
        except RuntimeError as e:
            if "No executable batch size found" in str(e):
                batch_size = 1
            else:
                raise

        if self.world_size > 1:
            # if multi-GPU, always take minimum over all selected batch sizes
            max_rnk_bs = torch.tensor([batch_size], device=self.device)
            gathered = self.accelerator.gather(max_rnk_bs).cpu().detach().numpy().tolist()
            batch_size = min(gathered)
            clear_torch_cache()
            return batch_size

        clear_torch_cache()
        return batch_size

    def tok_encode(self, string: str, left_truncate_len=None, add_special_tokens=None) -> List[int]:
        """"""
        if add_special_tokens is None:
            if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
                add_special_tokens = False or self.add_bos_token
            elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
                # TODO: investigate best practices for enc-dec models + special tokens
                add_special_tokens = True

        encoding = self.tokenizer.encode(string, add_special_tokens=add_special_tokens)

        # left-truncate the encoded context to be at most `left_truncate_len` tokens long
        if left_truncate_len:
            encoding = encoding[-left_truncate_len:]  # pylint: disable=E1130

        return encoding

    def tok_batch_encode(
        self,
        strings: List[str],
        padding_side: str = "left",
        left_truncate_len: int = None,
        truncation: bool = False,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # encode a batch of strings. converts to tensors and pads automatically, unlike tok_encode.
        old_padding_side = self.tokenizer.padding_side
        self.tokenizer.padding_side = padding_side

        if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
            add_special_tokens = False or self.add_bos_token
        elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
            add_special_tokens = True

        encoding = self.tokenizer(
            strings,
            truncation=truncation,
            padding="longest",
            return_tensors="pt",
            add_special_tokens=add_special_tokens,
        )
        if left_truncate_len:
            encoding["input_ids"] = encoding["input_ids"][:, -left_truncate_len:]  # pylint: disable=E1130
            encoding["attention_mask"] = encoding["attention_mask"][:, -left_truncate_len:]  # pylint: disable=E1130
        self.tokenizer.padding_side = old_padding_side

        return encoding["input_ids"], encoding["attention_mask"]

    def tok_decode(self, tokens, skip_special_tokens=True):
        if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
            return self.tokenizer.decode(tokens, skip_special_tokens=skip_special_tokens)
        elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
            return self.tokenizer.decode(tokens, skip_special_tokens=skip_special_tokens)

    def find_bucket(self, length):
        suitable_buckets = [b for b in self.buckets if b >= length]
        if len(suitable_buckets) == 0:
            eval_logger.error(f"The input_length={length} exceeds the maximum value in buckets={self.buckets}")
            eval_logger.error("Please add a higher value into the buckets list for this case.")
            exit(0)
        else:
            if self.last_bucket != suitable_buckets[0]:
                if hasattr(self.model, "clear_cache"):
                    self.model.clear_cache()  # clear HPU graph cache to avoid OOM
                self.last_bucket = suitable_buckets[0]
            return self.last_bucket

    def _model_call(self, inps, attn_mask=None, labels=None):
        """
        :param inps: torch.Tensor
            A torch tensor of shape [batch, (sequence_ctx + sequence_cont)] or of shape
            [batch, sequence_ctx]. the size of sequence may vary from call to call
        :param attn_mask: torch.Tensor, optional
            A torch tensor of shape [batch, (sequence_ctx + sequence_cont)]. Only passed
            (and must be passed) if self.AUTO_MODEL_CLASS is neural_compressor
        :param labels: torch.Tensor, optional
            A torch tensor of shape [batch, (sequence_ctx + sequence_cont)]. Only passed
            (and must be passed) if self.AUTO_MODEL_CLASS is neural_compressor.
            .transformers.AutoModelForSeq2SeqLM
        :return
            A torch tensor of shape [batch, sequence, vocab] with the
        logits returned from the model's decoder
        """
        with torch.no_grad():
            if attn_mask is not None or labels is not None:
                assert attn_mask is not None and labels is not None
                assert self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM
                if self.model_format == "onnx":
                    decoder_start_token_id = self._config.decoder_start_token_id
                    pad_token_id = self._config.pad_token_id
                    shifted_input_ids = labels.new_zeros(labels.shape)
                    shifted_input_ids[..., 1:] = labels[..., :-1].clone()
                    shifted_input_ids[..., 0] = decoder_start_token_id
                    shifted_input_ids.masked_fill_(shifted_input_ids == -100, pad_token_id)
                    return self.model(
                        inps,
                        attention_mask=attn_mask,
                        decoder_input_ids=shifted_input_ids,
                        labels=labels,
                    ).logits
                else:
                    output = self.model(input_ids=inps, attention_mask=attn_mask, labels=labels)
                    if isinstance(output, tuple):
                        output = output[0]
                    else:
                        output = output.logits
                    return output
            else:
                assert self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM
                if (
                    hasattr(self.model, "config")
                    and hasattr(self.model.config, "auto_map")
                    and "chatglm2" in self.model.config.auto_map["AutoConfig"]
                ):
                    input_bs, input_len = inps.shape
                    bos = torch.tensor([64790, 64792]).repeat(input_bs, 1)
                    inps = torch.cat((bos, inps), 1)
                if self.model_format == "onnx":
                    inputs_names = [input.name for input in self.model.model.get_inputs()]
                    if "position_ids" in inputs_names:
                        # model is exported with optimum >= 1.14.0 with new input 'position_ids'
                        input_shape = inps.shape
                        position_ids = (
                            torch.arange(0, input_shape[-1], dtype=torch.long).unsqueeze(0).view(-1, input_shape[-1])
                        )
                        output = self.model(
                            inps,
                            torch.ones(inps.shape, dtype=torch.int64),
                            position_ids,
                        ).logits
                    else:
                        output = self.model(inps, torch.ones(inps.shape, dtype=torch.int64)).logits
                else:
                    if self.pad_to_buckets:  # use buckets to pad inputs
                        bs, seq_length = inps.shape
                        padding_length = 0
                        bucket_length = self.find_bucket(seq_length)
                        padding_length = bucket_length - seq_length
                        inps = F.pad(inps, (0, padding_length), value=self.model.config.pad_token_id)
                    output = self.model(inps)
                    if isinstance(output, tuple):
                        output = output[0]
                    else:
                        output = output.logits
                    if self.pad_to_buckets and padding_length != 0:  # use buckets to pad inputs
                        output = output[:, :-padding_length, :]
                    if "hpu" in output.device.type:  # make sure return fp32 tensor for HPU, TODO: root cause
                        output = output.to(torch.float32)
                return output

    def _model_generate(self, context, max_length, stop, **generation_kwargs):
        # temperature = 0.0 if not set
        # if do_sample is false and temp==0.0:
        # remove temperature, as do_sample=False takes care of this
        # and we don't want a warning from HF
        generation_kwargs["temperature"] = generation_kwargs.get("temperature", 0.0)
        do_sample = generation_kwargs.get("do_sample", None)

        # The temperature has to be a strictly positive float -- if it is 0.0, use greedy decoding strategies
        if generation_kwargs.get("temperature") == 0.0 and do_sample is None:
            generation_kwargs["do_sample"] = do_sample = False

        if do_sample is False and generation_kwargs.get("temperature") == 0.0:
            generation_kwargs.pop("temperature")
        # build stopping criteria
        stopping_criteria = stop_sequences_criteria(self.tokenizer, stop, context.shape[1], context.shape[0])
        return self.model.generate(
            input_ids=context,
            max_length=max_length,
            stopping_criteria=stopping_criteria,
            pad_token_id=self.tokenizer.pad_token_id,
            use_cache=True,
            **generation_kwargs,
        )

    def _select_cont_toks(self, logits: torch.Tensor, contlen: int = None, inplen: int = None) -> torch.Tensor:
        if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
            assert contlen and inplen, "Must pass input len and cont. len to select scored logits for causal LM"
            # discard right-padding.
            # also discard the input/context tokens. we'll only score continuations.
            logits = logits[inplen - contlen : inplen]
        elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
            assert contlen and not inplen, "Selecting scored logits for Seq2SeqLM requires only cont. len"
            # only discard right-padding.
            # the logits input to this fn only contain decoder-side tokens.
            logits = logits[:contlen]

        return logits

    def loglikelihood_rolling(self, requests: List[Instance], disable_tqdm: bool = False) -> List[float]:
        loglikelihoods = []

        adaptive_batch_size = None
        if self.batch_size == "auto":
            # using rolling window with maximum context
            print("Passed argument batch_size = auto. Detecting largest batch size")
            batch_size = self._detect_batch_size()
            print(f"Determined Largest batch size: {batch_size}")
            adaptive_batch_size = batch_size

        for (string,) in tqdm([req.args for req in requests], disable=(disable_tqdm or (self.rank != 0))):
            rolling_token_windows = list(
                map(
                    utils.make_disjoint_window,
                    utils.get_rolling_token_windows(
                        token_list=self.tok_encode(string),
                        prefix_token=self.eot_token_id,
                        max_seq_len=self.max_length,
                        context_len=1,
                    ),
                )
            )

            # TODO: Right now,
            # we pass single EOT token to the Encoder and the full context to the decoder, in seq2seq case
            rolling_token_windows = [(None,) + x for x in rolling_token_windows]

            pad_amnt = 0
            if self.world_size > 1:
                # We pad out the external document-level iterator so the inner iterator doesn't hang
                mytensor = torch.tensor(len(rolling_token_windows), device=self.device)
                gathered = self.accelerator.gather(mytensor).cpu().detach().numpy().tolist()

                pad_amnt = max(gathered) - gathered[self.rank]
                if pad_amnt > 0:
                    rolling_token_windows += pad_amnt * [rolling_token_windows[0]]

            string_nll = self._loglikelihood_tokens(
                requests=rolling_token_windows,
                disable_tqdm=True,
                override_bs=adaptive_batch_size,
            )

            if (self.world_size > 1) and (pad_amnt > 0):
                string_nll = [x[0] for x in string_nll[:-pad_amnt]]
            else:
                # discard is_greedy
                string_nll = [x[0] for x in string_nll]

            string_nll = sum(string_nll)
            loglikelihoods.append(string_nll)

        return loglikelihoods

    def _batch_scheduler(self, pos, n_reordered_requests):
        sched = pos // int(len(n_reordered_requests) / self.batch_schedule)
        if sched in self.batch_sizes:
            return self.batch_sizes[sched]
        if (len(self.batch_sizes) > 1) and (self.batch_sizes[sched - 1] == self.max_batch_size):
            # if previous batch size is already maximal, skip recomputation
            self.batch_sizes[sched] = self.max_batch_size
            return self.batch_sizes[sched]
        print(f"Passed argument batch_size = auto:{self.batch_schedule}. Detecting largest batch size")
        self.batch_sizes[sched] = self._detect_batch_size(n_reordered_requests, pos)
        print(f"Determined largest batch size: {self.batch_sizes[sched]}")
        return self.batch_sizes[sched]

    def _loglikelihood_tokens(
        self,
        requests: List[Tuple[Tuple[str, str], List[int], List[int]]],
        disable_tqdm: bool = False,
        override_bs: int = None,
    ) -> List[Tuple[float, bool]]:
        # TODO:
        # implement some kind of efficient-request-middleware that lumps together requests with the same context
        res = []

        def _collate(req: Tuple[Tuple[str, str], List[int], List[int]]):
            """Defines the key for the sorted method."""
            # the negative sign on len(toks) sorts descending - this has a few advantages:
            # - time estimates will always be over not underestimates, which is more useful for planning
            # - to know the size of a batch when going through the list, you know the first one is always the batch
            #   padded context length. this is useful to simplify the batching logic and more importantly to make
            #   automatic adaptive batches much much easier to implement
            # - any OOMs will happen right away rather than near the end

            toks = req[1] + req[2]
            return -len(toks), tuple(toks)

        def _lookup_one_token_cont(req: Tuple[Tuple[str, str], List[int], List[int]]):
            """Defines the key to group and lookup one-token continuations."""
            # Use with group_by="contexts" (optional)"
            # allows for the creation of a lookup, so we can reuse logits in case of one-token continuations.
            # speeds up some multiple-choice tasks proportionally to the number of choices.
            # groups requests by context+continuation[:-1] and infer on one request/group.
            return req[-2] + req[-1][:-1]

        re_ord = Collator(
            requests,
            sort_fn=_collate,
            group_by=(
                "contexts" if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM and self.logits_cache else None
            ),
            group_fn=_lookup_one_token_cont,
        )

        # automatic (variable) batch size detection for vectorization
        # pull longest context sample from request
        n_reordered_requests = len(re_ord)
        batch_size = self.batch_size if self.batch_size != "auto" else override_bs if override_bs is not None else 0
        batch_fn = (
            self._batch_scheduler
            if self.batch_size == "auto" and n_reordered_requests > 0 and not override_bs
            else None
        )

        chunks = re_ord.get_batched(n=batch_size, batch_fn=batch_fn)
        pbar = tqdm(
            total=len(requests),
            disable=(disable_tqdm or (self.rank != 0)),
            desc="Running loglikelihood requests",
        )
        for chunk in chunks:
            inps = []
            cont_toks_list = []
            inplens = []

            conts = []
            encoder_attns = []

            padding_len_inp = None
            padding_len_cont = None
            # because vectorizing is annoying, we first convert each (context, continuation) pair to padded
            # tensors, then we pack them together into a batch, call the model, and then pick it all apart
            # again because vectorizing is annoying

            for _, context_enc, continuation_enc in chunk:
                # sanity check
                assert len(context_enc) > 0
                assert len(continuation_enc) > 0
                assert len(continuation_enc) <= self.max_length

                # how this all works (illustrated on a causal decoder-only setup):
                #          CTX      CONT
                # inp    0 1 2 3|4 5 6 7 8 9   <- last token is deleted by inp[:, :-1]
                # model  \               \
                # logits   1 2 3|4 5 6 7 8 9   <- the ctx half gets tossed out by the
                # cont_toks      4 5 6 7 8 9      [:, -len(continuation_enc):, :self.vocab_size] slice

                # when too long to fit in context, truncate from the left
                if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
                    inp = torch.tensor(
                        (context_enc + continuation_enc)[-(self.max_length + 1) :][:-1],
                        dtype=torch.long,
                        device=self.device,
                    )
                    (inplen,) = inp.shape
                elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
                    inp = torch.tensor(
                        (context_enc)[-self.max_length :],  # pylint: disable=E1130
                        dtype=torch.long,
                        device=self.device,
                    )
                    (inplen,) = inp.shape

                    # build encoder attn masks
                    encoder_attns.append(torch.ones_like(inp))

                    cont = torch.tensor(
                        (continuation_enc)[-self.max_length :],  # pylint: disable=E1130
                        # TODO: left-shift these?
                        # TODO: our code assumes we never end up truncating conts for either model type
                        dtype=torch.long,
                        device=self.device,
                    )
                    (contlen,) = cont.shape

                    conts.append(cont)

                    padding_len_cont = max(padding_len_cont, contlen) if padding_len_cont is not None else contlen

                padding_len_inp = max(padding_len_inp, inplen) if padding_len_inp is not None else inplen

                inps.append(inp)  # [1, inp_length]
                cont_toks_list.append(continuation_enc)
                inplens.append(inplen)

            # create encoder attn mask and batched conts, if seq2seq
            call_kwargs = {}
            if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
                batched_inps = pad_and_concat(padding_len_inp, inps, padding_side="right")  # [batch, padding_len_inp]
            elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
                # TODO: left-pad encoder inps and mask?
                batched_inps = pad_and_concat(padding_len_inp, inps)  # [batch, padding_len_inp]
                batched_conts = pad_and_concat(padding_len_cont, conts)  # [batch, padding_len_cont]
                batched_encoder_mask = pad_and_concat(padding_len_inp, encoder_attns)  # [batch, padding_len_inp]
                call_kwargs = {
                    "attn_mask": batched_encoder_mask,
                    "labels": batched_conts,
                }

            multi_logits = F.log_softmax(
                self._model_call(batched_inps, **call_kwargs), dim=-1
            )  # [batch, padding_length (inp or cont), vocab]

            for (request_str, ctx_tokens, _), logits, inplen, cont_toks in zip(
                chunk, multi_logits, inplens, cont_toks_list
            ):
                # Slice to original seq length
                contlen = len(cont_toks)
                # take only logits in the continuation
                # (discard context toks if decoder-only ; discard right-padding)
                # also discards + checks for "virtual tokens" in the causal LM's input window
                # from prompt/prefix tuning tokens, if applicable
                ctx_len = (
                    inplen + (logits.shape[0] - padding_len_inp)
                    if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM
                    else None
                )
                logits = self._select_cont_toks(logits, contlen=contlen, inplen=ctx_len)
                logits = logits.unsqueeze(0)  # [1, seq, vocab]

                # Check if per-token argmax is exactly equal to continuation
                greedy_tokens = logits.argmax(dim=-1)

                # check for one-token continuation cache hits.
                # noop in case group_by != "contexts" or no cache hit and returns the
                # original args. Otherwise, expands the logits batch dimension and yields each
                # batch along with matching continuation tokens and prompt strings.
                # logits -> [1, seq, vocab]
                for request_str, cont_toks, logits in re_ord.get_cache(
                    req_str=request_str,
                    cxt_toks=ctx_tokens,
                    cont_toks=cont_toks,
                    logits=logits,
                ):
                    cont_toks = torch.tensor(cont_toks, dtype=torch.long, device=self.device).unsqueeze(0)  # [1, seq]
                    max_equal = (greedy_tokens == cont_toks).all()

                    # Obtain log-probs at the corresponding continuation token indices
                    # last_token_slice = logits[:, -1, :].squeeze(0).tolist()
                    logits = torch.gather(logits, 2, cont_toks.unsqueeze(-1)).squeeze(-1)  # [1, seq]

                    # Answer: (log prob, is-exact-match)
                    answer = (float(logits.sum()), bool(max_equal))

                    res.append(answer)

                    self.cache_hook.add_partial("loglikelihood", request_str, answer)
                    pbar.update(1)

        pbar.close()

        return re_ord.get_original(res)

    def generate_until(self, requests: List[Instance], disable_tqdm: bool = False) -> List[str]:
        res = []

        def _collate(req: Tuple[str, dict]):
            """Defines the key for the sorted method."""
            # the negative sign on len(toks) sorts descending - this has a few advantages:
            # - time estimates will always be over not underestimates, which is more useful for planning
            # - to know the size of a batch when going through the list, you know the first one is always the batch
            #   padded context length. this is useful to simplify the batching logic and more importantly to make
            #   automatic adaptive batches much much easier to implement
            # - any OOMs will happen right away rather than near the end
            toks = self.tok_encode(req[0])
            return -len(toks), req[0]

        pbar = tqdm(
            total=len(requests),
            disable=(disable_tqdm or (self.rank != 0)),
            desc="Running generate_until requests",
        )
        adaptive_batch_size = None
        if self.batch_size == "auto":
            # using rolling window with maximum context
            print("Passed argument batch_size = auto. Detecting largest batch size")
            batch_size = self._detect_batch_size()
            print(f"Determined Largest batch size: {batch_size}")
            adaptive_batch_size = batch_size
        # for each different set of kwargs, we execute all requests, by batch.
        batch_size = (
            self.batch_size
            if self.batch_size != "auto"
            else adaptive_batch_size if adaptive_batch_size is not None else 0
        )
        batch_fn = self._batch_scheduler if self.batch_size == "auto" and not adaptive_batch_size else None

        # we group requests by their generation_kwargs,
        # so that we don't try to execute e.g. greedy sampling and temp=0.8 sampling
        # in the same batch.
        # group_fn=lambda x: x[1] -> x=(context, gen_kwargs)
        re_ords = Collator(
            [reg.args for reg in requests],
            sort_fn=_collate,
            group_by="gen_kwargs",
            group_fn=lambda x: x[1],
        )
        chunks = re_ords.get_batched(n=batch_size, batch_fn=batch_fn)
        for chunk in chunks:
            contexts, all_gen_kwargs = zip(*chunk)
            # we assume all gen kwargs in the batch are the same
            # this is safe to assume because the `grouper` object ensures it.
            gen_kwargs = all_gen_kwargs[0]
            # unpack our keyword arguments.
            until = None
            if isinstance(gen_kwargs, dict):
                kwargs = copy.deepcopy(gen_kwargs)  # edge case for repeats > 1
                if "until" in kwargs.keys():
                    until = kwargs.pop("until")
                    if isinstance(until, str):
                        until = [kwargs]
                    elif not isinstance(until, list):
                        raise ValueError(f"Expected `kwargs['until']` to be of type Union[str,list] but got {until}")
            else:
                raise ValueError(f"Expected `kwargs` to be of type `dict` but got {type(gen_kwargs)}")
            # add EOS token to stop sequences
            eos = self.tok_decode(self.eot_token_id, skip_special_tokens=False)
            if not until:
                until = [eos]
            else:
                until.append(eos)
            if "max_gen_toks" in kwargs.keys():
                max_gen_toks = kwargs.pop("max_gen_toks")
            else:
                max_gen_toks = self.max_gen_toks

            # set the max length in tokens of inputs ("context_enc")
            if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
                # max len for inputs = max length, minus room to generate the max new tokens
                max_ctx_len = self.max_length - max_gen_toks
            elif self.AUTO_MODEL_CLASS == transformers.AutoModelForSeq2SeqLM:
                # max len for inputs = encoder's whole max_length
                max_ctx_len = self.max_length

            # encode, pad, and truncate contexts for this batch
            context_enc, attn_masks = self.tok_batch_encode(
                contexts,
                left_truncate_len=max_ctx_len,
                truncation=self.truncation,
            )
            context_enc = context_enc.to(self.device)
            attn_masks = attn_masks.to(self.device)

            if "max_length" not in kwargs:
                kwargs["max_length"] = context_enc.shape[1] + max_gen_toks

            # perform batched generation
            cont = self._model_generate(
                context=context_enc,
                attention_mask=attn_masks,
                stop=until,
                **kwargs,
            )

            cont_toks_list = cont.tolist()
            for cont_toks, context in zip(cont_toks_list, contexts):
                # discard context + left-padding toks if using causal decoder-only LM
                if self.AUTO_MODEL_CLASS == transformers.AutoModelForCausalLM:
                    cont_toks = cont_toks[context_enc.shape[1] :]

                s = self.tok_decode(cont_toks)

                # use secondary stop seqs to cut off should-have-been-stopped content post-hoc
                for term in until:
                    if len(term) > 0:
                        # ignore '' separator,
                        # for seq2seq case where self.tok_decode(self.eot_token_id) = ''
                        s = s.split(term)[0]

                res.append(s)

                self.cache_hook.add_partial("generate_until", (context, gen_kwargs), s)
                pbar.update(1)
        # reorder this group of results back to original unsorted form
        res = re_ords.get_original(res)

        pbar.close()

        return res



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for bit packing."""


from typing import Callable, Dict, Tuple

import numba
import numpy as np

#  key: (bits, compress_bits), value: pack function
bit_packers: Dict[Tuple[int, int], Callable] = {}


def register_pack_func(orig_bits: int, compress_bits: int):
    """Register the pack function."""

    def decorator(func):
        bit_packers[(orig_bits, compress_bits)] = func
        return func

    return decorator


@register_pack_func(4, 32)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b4_c32(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=4 and compress_bits=32."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 7] & 0b1111) << 28)
            | ((raw_array[:, i * n_pack + 6] & 0b1111) << 24)
            | ((raw_array[:, i * n_pack + 5] & 0b1111) << 20)
            | ((raw_array[:, i * n_pack + 4] & 0b1111) << 16)
            | ((raw_array[:, i * n_pack + 3] & 0b1111) << 12)
            | ((raw_array[:, i * n_pack + 2] & 0b1111) << 8)
            | ((raw_array[:, i * n_pack + 1] & 0b1111) << 4)
            | (raw_array[:, i * n_pack] & 0b1111)
        )
    return packed_array


@register_pack_func(4, 16)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b4_c16(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=4 and compress_bits=16."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 3] & 0b1111) << 12)
            | ((raw_array[:, i * n_pack + 2] & 0b1111) << 8)
            | ((raw_array[:, i * n_pack + 1] & 0b1111) << 4)
            | (raw_array[:, i * n_pack] & 0b1111)
        )
    return packed_array


@register_pack_func(4, 8)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b4_c8(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=4 and compress_bits=8."""
    for i in range(new_in_features):
        packed_array[:, i] = ((raw_array[:, i * n_pack + 1] & 0b1111) << 4) | (raw_array[:, i * n_pack] & 0b1111)
    return packed_array


@register_pack_func(4, 64)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b4_c64(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=4 and compress_bits=64."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 15] & 0b1111) << 60)
            | ((raw_array[:, i * n_pack + 14] & 0b1111) << 56)
            | ((raw_array[:, i * n_pack + 13] & 0b1111) << 52)
            | ((raw_array[:, i * n_pack + 12] & 0b1111) << 48)
            | ((raw_array[:, i * n_pack + 11] & 0b1111) << 44)
            | ((raw_array[:, i * n_pack + 10] & 0b1111) << 40)
            | ((raw_array[:, i * n_pack + 9] & 0b1111) << 36)
            | ((raw_array[:, i * n_pack + 8] & 0b1111) << 32)
            | ((raw_array[:, i * n_pack + 7] & 0b1111) << 28)
            | ((raw_array[:, i * n_pack + 6] & 0b1111) << 24)
            | ((raw_array[:, i * n_pack + 5] & 0b1111) << 20)
            | ((raw_array[:, i * n_pack + 4] & 0b1111) << 16)
            | ((raw_array[:, i * n_pack + 3] & 0b1111) << 12)
            | ((raw_array[:, i * n_pack + 2] & 0b1111) << 8)
            | ((raw_array[:, i * n_pack + 1] & 0b1111) << 4)
            | (raw_array[:, i * n_pack] & 0b1111)
        )
    return packed_array


@register_pack_func(8, 32)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b8_c32(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=8 and compress_bits=32."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 3] & 0b11111111) << 24)
            | ((raw_array[:, i * n_pack + 2] & 0b11111111) << 16)
            | ((raw_array[:, i * n_pack + 1] & 0b11111111) << 8)
            | (raw_array[:, i * n_pack] & 0b11111111)
        )
    return packed_array


@register_pack_func(8, 16)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b8_c16(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=8 and compress_bits=16."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 3] & 0b11111111) << 24)
            | ((raw_array[:, i * n_pack + 2] & 0b11111111) << 16)
            | ((raw_array[:, i * n_pack + 1] & 0b11111111) << 8)
            | (raw_array[:, i * n_pack] & 0b11111111)
        )
    return packed_array


@register_pack_func(8, 8)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b8_c8(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=8 and compress_bits=8."""
    for i in range(new_in_features):
        packed_array[:, i] = raw_array[:, i * n_pack] & 0b11111111
    return packed_array


@register_pack_func(8, 64)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b8_c64(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=8 and compress_bits=64."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 7] & 0b11111111) << 56)
            | ((raw_array[:, i * n_pack + 6] & 0b11111111) << 48)
            | ((raw_array[:, i * n_pack + 5] & 0b11111111) << 40)
            | ((raw_array[:, i * n_pack + 4] & 0b11111111) << 32)
            | ((raw_array[:, i * n_pack + 3] & 0b11111111) << 24)
            | ((raw_array[:, i * n_pack + 2] & 0b11111111) << 16)
            | ((raw_array[:, i * n_pack + 1] & 0b11111111) << 8)
            | (raw_array[:, i * n_pack] & 0b11111111)
        )
    return packed_array


@register_pack_func(2, 32)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b2_c32(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=2 and compress_bits=32."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 15] & 0b11) << 30)
            | ((raw_array[:, i * n_pack + 14] & 0b11) << 28)
            | ((raw_array[:, i * n_pack + 13] & 0b11) << 26)
            | ((raw_array[:, i * n_pack + 12] & 0b11) << 24)
            | ((raw_array[:, i * n_pack + 11] & 0b11) << 22)
            | ((raw_array[:, i * n_pack + 10] & 0b11) << 20)
            | ((raw_array[:, i * n_pack + 9] & 0b11) << 18)
            | ((raw_array[:, i * n_pack + 8] & 0b11) << 16)
            | ((raw_array[:, i * n_pack + 7] & 0b11) << 14)
            | ((raw_array[:, i * n_pack + 6] & 0b11) << 12)
            | ((raw_array[:, i * n_pack + 5] & 0b11) << 10)
            | ((raw_array[:, i * n_pack + 4] & 0b11) << 8)
            | ((raw_array[:, i * n_pack + 3] & 0b11) << 6)
            | ((raw_array[:, i * n_pack + 2] & 0b11) << 4)
            | ((raw_array[:, i * n_pack + 1] & 0b11) << 2)
            | (raw_array[:, i * n_pack] & 0b11)
        )
    return packed_array


@register_pack_func(2, 16)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b2_c16(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=2 and compress_bits=16."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 7] & 0b11) << 14)
            | ((raw_array[:, i * n_pack + 6] & 0b11) << 12)
            | ((raw_array[:, i * n_pack + 5] & 0b11) << 10)
            | ((raw_array[:, i * n_pack + 4] & 0b11) << 8)
            | ((raw_array[:, i * n_pack + 3] & 0b11) << 6)
            | ((raw_array[:, i * n_pack + 2] & 0b11) << 4)
            | ((raw_array[:, i * n_pack + 1] & 0b11) << 2)
            | (raw_array[:, i * n_pack] & 0b11)
        )
    return packed_array


@register_pack_func(2, 8)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b2_c8(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=2 and compress_bits=8."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 3] & 0b11) << 6)
            | ((raw_array[:, i * n_pack + 2] & 0b11) << 4)
            | ((raw_array[:, i * n_pack + 1] & 0b11) << 2)
            | (raw_array[:, i * n_pack] & 0b11)
        )
    return packed_array


@register_pack_func(2, 64)
@numba.jit(nopython=True, parallel=True)
def pack_array_with_numba_b2_c64(
    raw_array: np.ndarray, packed_array: np.ndarray, n_pack: int, new_in_features: int
) -> np.ndarray:
    """Pack the array with numba when bits=2 and compress_bits=64."""
    for i in range(new_in_features):
        packed_array[:, i] = (
            ((raw_array[:, i * n_pack + 31] & 0b11) << 62)
            | ((raw_array[:, i * n_pack + 30] & 0b11) << 60)
            | ((raw_array[:, i * n_pack + 29] & 0b11) << 58)
            | ((raw_array[:, i * n_pack + 28] & 0b11) << 56)
            | ((raw_array[:, i * n_pack + 27] & 0b11) << 54)
            | ((raw_array[:, i * n_pack + 26] & 0b11) << 52)
            | ((raw_array[:, i * n_pack + 25] & 0b11) << 50)
            | ((raw_array[:, i * n_pack + 24] & 0b11) << 48)
            | ((raw_array[:, i * n_pack + 23] & 0b11) << 46)
            | ((raw_array[:, i * n_pack + 22] & 0b11) << 44)
            | ((raw_array[:, i * n_pack + 21] & 0b11) << 42)
            | ((raw_array[:, i * n_pack + 20] & 0b11) << 40)
            | ((raw_array[:, i * n_pack + 19] & 0b11) << 38)
            | ((raw_array[:, i * n_pack + 18] & 0b11) << 36)
            | ((raw_array[:, i * n_pack + 17] & 0b11) << 34)
            | ((raw_array[:, i * n_pack + 16] & 0b11) << 32)
            | ((raw_array[:, i * n_pack + 15] & 0b11) << 30)
            | ((raw_array[:, i * n_pack + 14] & 0b11) << 28)
            | ((raw_array[:, i * n_pack + 13] & 0b11) << 26)
            | ((raw_array[:, i * n_pack + 12] & 0b11) << 24)
            | ((raw_array[:, i * n_pack + 11] & 0b11) << 22)
            | ((raw_array[:, i * n_pack + 10] & 0b11) << 20)
            | ((raw_array[:, i * n_pack + 9] & 0b11) << 18)
            | ((raw_array[:, i * n_pack + 8] & 0b11) << 16)
            | ((raw_array[:, i * n_pack + 7] & 0b11) << 14)
            | ((raw_array[:, i * n_pack + 6] & 0b11) << 12)
            | ((raw_array[:, i * n_pack + 5] & 0b11) << 10)
            | ((raw_array[:, i * n_pack + 4] & 0b11) << 8)
            | ((raw_array[:, i * n_pack + 3] & 0b11) << 6)
            | ((raw_array[:, i * n_pack + 2] & 0b11) << 4)
            | ((raw_array[:, i * n_pack + 1] & 0b11) << 2)
            | (raw_array[:, i * n_pack] & 0b11)
        )
    return packed_array



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random
import time

import evaluate  # pylint: disable=E0401
import numpy as np
import torch
from tqdm import tqdm

from .hf_datasets.cnn_dailymail import CNNDAILYMAIL, postprocess_text

seed = 9973
random.seed(seed)
np.random.seed(seed)
torch.manual_seed(seed)


def summarization_evaluate(model, tokenizer_name=None, task="cnn_dailymail", batch_size=1, limit=None):
    generate_kwargs = {
        "max_new_tokens": 128,
        "min_new_tokens": 30,
        "num_beams": 4,
        "early_stopping": True,
    }
    metric = evaluate.load("rouge")
    val_dataset = CNNDAILYMAIL(tokenizer_name, device="cpu", calib=False, num_samples=limit)
    sources = val_dataset.sources
    targets = val_dataset.targets
    tokenizer = val_dataset.tokenizer
    preds = []
    predictions = []
    ground_truths = []
    max_len = 1919
    with torch.inference_mode(), torch.no_grad():
        for i in tqdm(range(0, len(sources), batch_size)):
            input_batch = tokenizer.batch_encode_plus(
                sources[i : i + batch_size],
                return_tensors="pt",
                padding=True,
                truncation=True,
                max_length=max_len,
            )
            input_ids = input_batch.input_ids
            input_lens = input_ids.shape[-1]
            t0 = time.time()
            out_tokens = model.generate(**input_batch, **generate_kwargs, pad_token_id=tokenizer.pad_token_id)
            t1 = time.time()
            print("Inference time: {}".format(round(t1 - t0, 3)))
            print("Seq len: {}".format(input_lens))
            if "t5" in tokenizer_name:
                pred = out_tokens
                print("Out len: {}".format(out_tokens.shape[-1]))
            else:
                pred = out_tokens[:, input_lens:]
                print("Out len: {}".format(out_tokens.shape[-1] - input_lens))
            pred_batch = tokenizer.batch_decode(pred, skip_special_tokens=True)
            targ_batch = targets[i : i + batch_size]
            preds, targs = postprocess_text(pred_batch, targ_batch)
            predictions.extend(preds)
            ground_truths.extend(targs)

            if limit is not None and i == int(limit / batch_size):
                break

    result = metric.compute(
        predictions=predictions, references=ground_truths, use_stemmer=False
    )  # , tokenizer=tokenizer)
    result = {k: round(v * 100, 4) for k, v in result.items()}

    print(result)
    return result



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch constants."""

# double quant params
DOUBLE_QUANT_CONFIGS = {
    "BNB_NF4": {
        "dtype": "nf4",
        "bits": 4,
        "group_size": 32,
        "use_double_quant": True,
        "double_quant_bits": 8,
        "double_quant_dtype": "int",
        "double_quant_use_sym": False,
        "double_quant_group_size": 256,
    },
    # TODO: (Xin) current implementation is not the same as GGML.
    # GGML is using double_quant_bits to quantize zero points
    "GGML_TYPE_Q4_K": {
        "dtype": "int",
        "bits": 4,
        "use_sym": False,
        "group_size": 32,
        "use_double_quant": True,
        "double_quant_bits": 6,
        "double_quant_dtype": "int",
        "double_quant_use_sym": True,
        "double_quant_group_size": 8,
    },
}

# Setting priorities for algorithms, a higher number indicates a higher priority.
PRIORITY_GPTQ = 90
PRIORITY_RTN = 80
PRIORITY_HQQ = 75
PRIORITY_AWQ = 70
PRIORITY_TEQ = 60
PRIORITY_AUTOROUND = 50


PT2E_STATIC_QUANT = "pt2e_static_quant"
PT2E_DYNAMIC_QUANT = "pt2e_dynamic_quant"


# load format name
from enum import Enum


class SaveLoadFormat(Enum):
    """Load API formats for different usages."""

    DEFAULT = "default"
    HUGGINGFACE = "huggingface"
    VLLM = "vllm"


LM_HEAD_NAMES = [".*lm_head", ".*output_layer", ".*embed_out"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MX quantization utils."""

from enum import Enum, IntEnum

import torch

FP32_EXPONENT_BIAS = 127
FP32_MIN_NORMAL = 2 ** (-FP32_EXPONENT_BIAS + 1)


class ElemFormat(Enum):
    """Element format."""

    int8 = 1
    int4 = 2
    int2 = 3
    fp8_e5m2 = 4
    fp8_e4m3 = 5
    fp6_e3m2 = 6
    fp6_e2m3 = 7
    fp4 = 8
    fp4_e2m1 = 8
    float16 = 9
    fp16 = 9
    bfloat16 = 10
    bf16 = 10

    @staticmethod
    def from_str(s):
        """Get element format with str."""
        assert s is not None, "String elem_format == None"
        s = s.lower()
        if hasattr(ElemFormat, s):
            return getattr(ElemFormat, s)
        else:
            raise Exception("Undefined elem format", s)

    @staticmethod
    def is_bf(s):
        """Whether the format is brain floating-point format."""
        if isinstance(s, str):
            assert s is not None, "String elem_format == None"
            s = s.lower()
            if hasattr(ElemFormat, s):
                return getattr(ElemFormat, s).value == 10
        elif isinstance(s, int):
            return s == 10

        raise Exception("Undefined elem format", s)

    @staticmethod
    def is_fp(s):
        """Whether the format is floating-point format."""
        if isinstance(s, str):
            assert s is not None, "String elem_format == None"
            s = s.lower()
            if hasattr(ElemFormat, s):
                return 4 <= getattr(ElemFormat, s).value <= 9
        elif isinstance(s, int):
            return 4 <= s <= 9

        raise Exception("Undefined elem format", s)

    @staticmethod
    def is_int(s):
        """Whether the format is integer format."""
        if isinstance(s, str):
            assert s is not None, "String elem_format == None"
            s = s.lower()
            if hasattr(ElemFormat, s):
                return 1 <= getattr(ElemFormat, s).value <= 3
        elif isinstance(s, int):
            return 1 <= s <= 3

        raise Exception("Undefined elem format", s)


class RoundingMode(IntEnum):
    """Rounding mode."""

    nearest = 0
    floor = 1
    even = 2

    @staticmethod
    def string_enums():
        """Rounding mode names."""
        return [s.name for s in list(RoundingMode)]


def _get_min_norm(ebits):
    """Valid for all float formats."""
    emin = 2 - (2 ** (ebits - 1))
    return 0 if ebits == 0 else 2**emin


def _get_max_norm(ebits, mbits):
    """Valid only for floats that define NaN."""
    assert ebits >= 5, "invalid for floats that don't define NaN"
    emax = 0 if ebits == 0 else 2 ** (ebits - 1) - 1
    return 2**emax * float(2 ** (mbits - 1) - 1) / 2 ** (mbits - 2)


_FORMAT_CACHE = {}


def _get_format_params(fmt):
    """Get parameters of the format.

    Allowed formats:
    - intX:         2 <= X <= 32, assume sign-magnitude, 1.xxx representation
    - floatX/fpX:   16 <= X <= 28, assume top exp is used for NaN/Inf
    - bfloatX/bfX:  9 <= X <= 32
    - fp4,                  no NaN/Inf
    - fp6_e3m2/e2m3,        no NaN/Inf
    - fp8_e4m3/e5m2,        e5m2 normal NaN/Inf, e4m3 special behavior

    Args:
        fmt (str od ElemFormat): format

    Returns:
      ebits: exponent bits
      mbits: mantissa bits: includes sign and implicit bits
      emax: max normal exponent
      max_norm: max normal number
      min_norm: min normal number
    """
    if type(fmt) is str:
        fmt = ElemFormat.from_str(fmt)

    if fmt in _FORMAT_CACHE:
        return _FORMAT_CACHE[fmt]

    if fmt == ElemFormat.int8:
        ebits, mbits = 0, 8
        emax = 0
    elif fmt == ElemFormat.int4:
        ebits, mbits = 0, 4
        emax = 0
    elif fmt == ElemFormat.int2:
        ebits, mbits = 0, 2
        emax = 0
    elif fmt == ElemFormat.fp8_e5m2:
        ebits, mbits = 5, 4
        emax = 2 ** (ebits - 1) - 1
    elif fmt == ElemFormat.fp8_e4m3:
        ebits, mbits = 4, 5
        emax = 2 ** (ebits - 1)
    elif fmt == ElemFormat.fp6_e3m2:
        ebits, mbits = 3, 4
        emax = 2 ** (ebits - 1)
    elif fmt == ElemFormat.fp6_e2m3:
        ebits, mbits = 2, 5
        emax = 2 ** (ebits - 1)
    elif fmt == ElemFormat.fp4:
        ebits, mbits = 2, 3
        emax = 2 ** (ebits - 1)
    elif fmt == ElemFormat.float16:
        ebits, mbits = 5, 12
        emax = 2 ** (ebits - 1) - 1
    elif fmt == ElemFormat.bfloat16:
        ebits, mbits = 8, 9
        emax = 2 ** (ebits - 1) - 1
    else:
        raise Exception("Unknown element format %s" % fmt)

    if fmt != ElemFormat.fp8_e4m3:
        max_norm = 2**emax * float(2 ** (mbits - 1) - 1) / 2 ** (mbits - 2)
    else:
        max_norm = 2**emax * 1.75  # FP8 has custom max_norm

    min_norm = _get_min_norm(ebits)

    _FORMAT_CACHE[fmt] = (ebits, mbits, emax, max_norm, min_norm)

    return ebits, mbits, emax, max_norm, min_norm


# Never explicitly compute 2**(-exp) since subnorm numbers have
# exponents smaller than -126
def _safe_lshift(x, bits, exp):
    if exp is None:
        return x * (2**bits)
    else:
        return x / (2**exp) * (2**bits)


def _safe_rshift(x, bits, exp):
    if exp is None:
        return x / (2**bits)
    else:
        return x / (2**bits) * (2**exp)


def _round_mantissa(A, bits, round, clamp=False):
    """Rounds mantissa to nearest bits depending on the rounding method 'round'.

    Args:
        A (torch.Tensor): input tensor
        bits (int): bit number of mantissa
        round (str): rounding method
                     "floor" rounds to the floor
                     "nearest" rounds to ceil or floor, whichever is nearest
        clamp (bool, optional): Whether do clip. Defaults to False.

    Returns:
        torch.Tensor: tensor with mantissas rounded
    """
    if round == "dither":
        rand_A = torch.rand_like(A, requires_grad=False)
        A = torch.sign(A) * torch.floor(torch.abs(A) + rand_A)
    elif round == "floor":
        A = torch.sign(A) * torch.floor(torch.abs(A))
    elif round == "nearest":
        A = torch.sign(A) * torch.floor(torch.abs(A) + 0.5)
    elif round == "even":
        absA = torch.abs(A)
        # find 0.5, 2.5, 4.5 ...
        maskA = ((absA - 0.5) % 2 == torch.zeros_like(A)).type(A.dtype)
        A = torch.sign(A) * (torch.floor(absA + 0.5) - maskA)
    else:
        raise Exception("Unrecognized round method %s" % (round))

    # Clip values that cannot be expressed by the specified number of bits
    if clamp:
        max_mantissa = 2 ** (bits - 1) - 1
        A = torch.clamp(A, -max_mantissa, max_mantissa)
    return A


def _shared_exponents(A, method="max", axes=None, ebits=0):
    """Get shared exponents for the passed matrix A.

    Args:
        A (torch.Tensor): Input tensor
        method (str, optional): Exponent selection method.
                                "max" uses the max absolute value.
                                "none" uses an exponent for each value (i.e., no sharing)
                                Defaults to "max".
        axes (list(int), optional): list of integers which specifies the axes across which
                                    shared exponents are calculated. Defaults to None.
        ebits (int, optional): bit number of the shared exponents. Defaults to 0.

    Returns:
        shared_exp (torch.Tensor): Tensor of shared exponents
    """
    if method == "max":
        if axes is None:
            shared_exp = torch.max(torch.abs(A))
        else:
            shared_exp = A
            for axis in axes:
                shared_exp, _ = torch.max(torch.abs(shared_exp), dim=axis, keepdim=True)
    elif method == "none":
        shared_exp = torch.abs(A)
    else:
        raise Exception("Unrecognized shared exponent selection method %s" % (method))

    # log2(shared_exp) and truncate to integer
    shared_exp = torch.floor(torch.log2(shared_exp + FP32_MIN_NORMAL * (shared_exp == 0).type(shared_exp.dtype)))

    # Restrict to [-emax, emax] range
    if ebits > 0:
        emax = 2 ** (ebits - 1) - 1
        # shared_exp = torch.clamp(shared_exp, -emax, emax)
        # Overflow to Inf
        shared_exp[shared_exp > emax] = float("NaN")
        # Underflows are set to -127 which causes them to be
        # flushed to 0 later
        shared_exp[shared_exp < -emax] = -emax

    return shared_exp


def _reshape_to_blocks(A, axes, block_size):
    if axes is None:
        raise Exception("axes required in order to determine which " "dimension to apply block size to")
    if block_size == 0:
        raise Exception("block_size == 0 in _reshape_to_blocks")

    # Fix axes to be positive and sort them
    axes = [(x + len(A.shape) if x < 0 else x) for x in axes]
    assert all(x >= 0 for x in axes)
    axes = sorted(axes)

    # Add extra dimension for tiles
    for i in range(len(axes)):
        axes[i] += i  # Shift axes due to added dimensions
        A = torch.unsqueeze(A, dim=axes[i] + 1)

    # Pad to block_size
    orig_shape = A.size()
    pad = []
    for i in range(len(orig_shape)):
        pad += [0, 0]

    do_padding = False
    for axis in axes:
        pre_pad_size = orig_shape[axis]
        if isinstance(pre_pad_size, torch.Tensor):
            pre_pad_size = int(pre_pad_size.value)
        # Don't pad if the axis is short enough to fit inside one tile
        if pre_pad_size % block_size == 0:
            pad[2 * axis] = 0
        else:
            pad[2 * axis] = block_size - pre_pad_size % block_size
            do_padding = True

    if do_padding:
        pad = list(reversed(pad))
        A = torch.nn.functional.pad(A, pad, mode="constant")

    def _reshape(shape, reshape_block_size):
        for axis in axes:
            # Reshape to tiles if axis length > reshape_block_size
            if shape[axis] >= reshape_block_size:
                assert shape[axis] % reshape_block_size == 0
                shape[axis + 1] = reshape_block_size
                shape[axis] = shape[axis] // reshape_block_size
            # Otherwise preserve length and insert a 1 into the shape
            else:
                shape[axis + 1] = shape[axis]
                shape[axis] = 1
        return shape

    # Reshape to tiles
    padded_shape = A.size()
    reshape = _reshape(list(padded_shape), block_size)

    A = A.view(reshape)
    return A, axes, orig_shape, padded_shape


def _undo_reshape_to_blocks(A, padded_shape, orig_shape, axes):
    # Undo tile reshaping
    A = A.view(padded_shape)
    # Undo padding
    if not list(padded_shape) == list(orig_shape):
        slices = [slice(0, x) for x in orig_shape]
        A = A[slices]
    for axis in reversed(axes):
        # Remove extra dimension
        A = torch.squeeze(A, dim=axis + 1)
    return A


def _quantize_elemwise_core(A, bits, exp_bits, max_norm, round="nearest", saturate_normals=False, allow_denorm=True):
    """Core function used for element-wise quantization.

    Args:
        A (torch.Tensor): tensor to be quantized
        bits (int): number of mantissa bits. Includes sign bit and implicit one for floats
        exp_bits (int): number of exponent bits, 0 for ints
        max_norm (float): largest representable normal number
        round (str, optional): rounding mode: (floor, nearest, even). Defaults to "nearest".
        saturate_normals (bool, optional): whether clip normal numbers that exceed max norm.
                                           Must be True for correct MX conversion. Defaults to False.
        allow_denorm (bool, optional): if False, flush denorm numbers in the elem_format to zero. Defaults to True.

    Returns:
        torch.Tensor: tensor that has been quantized
    """
    # Flush values < min_norm to zero if denorms are not allowed
    if not allow_denorm and exp_bits > 0:
        min_norm = _get_min_norm(exp_bits)
        out = (torch.abs(A) >= min_norm).type(A.dtype) * A
    else:
        out = A

    if exp_bits != 0:
        private_exp = torch.floor(torch.log2(torch.abs(A) + (A == 0).type(A.dtype)))

        # The minimum representable exponent for 8 exp bits is -126
        min_exp = -(2 ** (exp_bits - 1)) + 2
        private_exp = private_exp.clip(min=min_exp)
    else:
        private_exp = None

    # Scale up so appropriate number of bits are in the integer portion of the number
    out = _safe_lshift(out, bits - 2, private_exp)

    out = _round_mantissa(out, bits, round, clamp=False)

    # Undo scaling
    out = _safe_rshift(out, bits - 2, private_exp)

    # Set values > max_norm to Inf if desired, else clamp them
    if saturate_normals or exp_bits == 0:
        out = torch.clamp(out, min=-max_norm, max=max_norm)
    else:
        out = torch.where((torch.abs(out) > max_norm), torch.sign(out) * float("Inf"), out)

    # handle Inf/NaN
    out[A == float("Inf")] = float("Inf")
    out[A == -float("Inf")] = -float("Inf")
    out[A == float("NaN")] = float("NaN")

    return out


def _quantize_fp(A, exp_bits=None, mantissa_bits=None, round="nearest", allow_denorm=True):
    """Quantize values to IEEE fpX format..

    The format defines NaN/Inf and subnorm numbers in the same way as FP32 and FP16.

    Args:
        A (torch.Tensor): a tensor that needs to be quantized
        exp_bits (int, optional): number of bits used to store exponent. Defaults to None.
        mantissa_bits (int, optional): number of bits used to store mantissa.
                                       Not including sign or implicit 1. Defaults to None.
        round (str, optional): rounding mode, (floor, nearest, even). Defaults to "nearest".
        allow_denorm (bool, optional): allow denorm numbers to exist. Defaults to True.

    Returns:
        torch.Tensor:  tensor that has been quantized
    """
    # Shortcut for no quantization
    if exp_bits is None or mantissa_bits is None:
        return A

    max_norm = _get_max_norm(exp_bits, mantissa_bits + 2)

    output = _quantize_elemwise_core(
        A, bits=mantissa_bits + 2, exp_bits=exp_bits, max_norm=max_norm, round=round, allow_denorm=allow_denorm
    )

    return output


def _quantize_bfloat(A, bfloat, round="nearest", allow_denorm=True):
    """Quantize values to bfloatX format.

    Args:
        A (torch.Tensor): a tensor that needs to be quantized
        bfloat (int): total number of bits for bfloatX format.
                      Includes 1 sign, 8 exp bits, and variable mantissa bits. Must be >= 9.
        round (str, optional): rounding mode, (floor, nearest, even). Defaults to "nearest".
        allow_denorm (bool, optional): allow denorm numbers to exist. Defaults to True.

    Returns:
        torch.Tensor:  tensor that has been quantized
    """
    # Shortcut for no quantization
    if bfloat == 0 or bfloat == 32:
        return A

    max_norm = _get_max_norm(8, bfloat - 7)

    return _quantize_elemwise_core(
        A, bits=bfloat - 7, exp_bits=8, max_norm=max_norm, round=round, allow_denorm=allow_denorm
    )


def quantize_elemwise_op(A, mx_specs):
    """A function used for element-wise quantization with mx_specs.

    Args:
        A (torch.Tensor): a tensor that needs to be quantized
        mx_specs (dict): dictionary to specify mx_specs

    Returns:
        torch.Tensor:  tensor that has been quantized
    """
    if mx_specs is None:
        return A

    out_dtype = mx_specs.out_dtype
    round = mx_specs.round_method
    elem_format = ElemFormat.from_str(out_dtype)
    ebits, mbits, _, _, _ = _get_format_params(elem_format)
    if ElemFormat.is_bf(out_dtype):
        A = _quantize_bfloat(A, bfloat=ebits + mbits - 1, round=round, allow_denorm=True)
    elif ElemFormat.is_fp(out_dtype):
        A = _quantize_fp(A, exp_bits=5, mantissa_bits=ebits + mbits - 1 - 6, round=round, allow_denorm=True)
    else:
        raise ValueError("Cannot set {} for output dtype.".format(out_dtype))
    return A


def _quantize_mx(
    A,
    scale_bits,
    elem_format,  # can be None for no quantization
    shared_exp_method="max",
    axes=None,
    block_size=32,
    round="nearest",
    flush_fp32_subnorms=False,
):
    """Function used for MX* quantization."""
    # Shortcut for no quantization
    if elem_format is None:
        return A

    assert scale_bits > 0

    # Make sure axes is a list of non-negative numbers
    axes = [axes] if type(axes) == int else axes
    axes = [x + A.ndim if x < 0 else x for x in axes]

    ebits, mbits, emax, max_norm, _ = _get_format_params(elem_format)

    # Perform tiling to the hardware vector size
    A, axes, orig_shape, padded_shape = _reshape_to_blocks(A, axes, block_size)

    ####################
    # Quantize
    ####################
    shared_exp_axes = [x + 1 for x in axes]

    # Get shared exponents
    shared_exp = _shared_exponents(
        A,
        method=shared_exp_method,
        axes=shared_exp_axes,
        ebits=0,
    )

    # Flush subnormal FP32 inputs to zero
    if flush_fp32_subnorms:
        A = A * (shared_exp > -FP32_EXPONENT_BIAS).type(A.dtype)

    # Offset the max exponent by the largest representable exponent
    # in the element data format
    shared_exp = shared_exp - emax

    scale_emax = 2 ** (scale_bits - 1) - 1
    shared_exp[shared_exp > scale_emax] = float("NaN")
    shared_exp[shared_exp < -scale_emax] = -scale_emax

    A = A / (2**shared_exp)

    A = _quantize_elemwise_core(A, mbits, ebits, max_norm, round=round, allow_denorm=True, saturate_normals=True)

    A = A * (2**shared_exp)

    # Undo tile reshaping
    A = _undo_reshape_to_blocks(A, padded_shape, orig_shape, axes)

    return A


def quantize_mx_op(
    A: torch.Tensor,
    elem_format: str,
    round: str,
    block_size: int,
    scale_bits=8,
    axes=None,
    expand_and_reshape=False,
):
    """Quantize tensor to MX data type."""
    if elem_format is None:
        return A
    elif type(elem_format) is str:
        elem_format = ElemFormat.from_str(elem_format)

    return _quantize_mx(
        A,
        scale_bits,
        elem_format,
        block_size=block_size,
        axes=axes,
        round=round,
        shared_exp_method="max",
        flush_fp32_subnorms=False,
    )
